from .aggregation_node import AggregationNode
from .cache_prefetch_node import CachePrefetchNode
from .component_iteration_node import ComponentIterationNode
from .exec_context import ExecContext
from .exec_node import ExecNode
from .expr_eval_node import ExprEvalNode
from .in_memory_data_node import InMemoryDataNode
from .sql_scan_node import SqlScanNode
from .media_validation_node import MediaValidationNode


from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple
import copy

import sqlalchemy as sql

from .expr import Expr
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.exceptions as excs
import pixeltable.catalog as catalog
import pixeltable.type_system as ts


class InlineDict(Expr):
    """
    Dictionary 'literal' which can use Exprs as values.
    """
    def __init__(self, d: Dict):
        from .inline_array import InlineArray
        super().__init__(ts.JsonType())  # we need to call this in order to populate self.components
        # dict_items contains
        # - for Expr fields: (key, index into components, None)
        # - for non-Expr fields: (key, -1, value)
        self.dict_items: List[Tuple[str, int, Any]] = []
        for key, val in d.items():
            if not isinstance(key, str):
                raise excs.Error(f'Dictionary requires string keys, {key} has type {type(key)}')
            val = copy.deepcopy(val)
            if isinstance(val, dict):
                val = InlineDict(val)
            if isinstance(val, list) or isinstance(val, tuple):
                val = InlineArray(tuple(val), force_json=True)
            if isinstance(val, Expr):
                self.dict_items.append((key, len(self.components), None))
                self.components.append(val)
            else:
                self.dict_items.append((key, -1, val))

        self.type_spec: Optional[Dict[str, ts.ColumnType]] = {}
        for key, idx, _ in self.dict_items:
            if idx == -1:
                # TODO: implement type inference for values
                self.type_spec = None
                break
            self.type_spec[key] = self.components[idx].col_type
        self.col_type = ts.JsonType(self.type_spec)

        self.id = self._create_id()

    def __str__(self) -> str:
        item_strs: List[str] = []
        i = 0
        def print_val(val: Any) -> str:
            if isinstance(val, str):
                return f"'{val}'"
            return str(val)
        for key, idx, val in self.dict_items:
            if idx != -1:
                item_strs.append(f"'{key}': {str(self.components[i])}")
                i += 1
            else:
                item_strs.append(f"'{key}': {print_val(val)}")
        return '{' + ', '.join(item_strs) + '}'

    def _equals(self, other: InlineDict) -> bool:
        return self.dict_items == other.dict_items

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('dict_items', self.dict_items)]

    def to_dict(self) -> Dict[str, Any]:
        """Return the original dict used to construct this"""
        return {key: val if idx == -1 else self.components[idx] for key, idx, val in self.dict_items}

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        result = {}
        for key, idx, val in self.dict_items:
            assert isinstance(key, str)
            if idx >= 0:
                result[key] = data_row[self.components[idx].slot_idx]
            else:
                result[key] = copy.deepcopy(val)
        data_row[self.slot_idx] = result

    def _as_dict(self) -> Dict:
        return {'dict_items': self.dict_items, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'dict_items' in d
        arg: Dict[str, Any] = {}
        for key, idx, val in d['dict_items']:
            if idx >= 0:
                arg[key] = components[idx]
            else:
                arg[key] = val
        return cls(arg)



from typing import List, Optional, Tuple, Iterable, Set
from uuid import UUID
import logging
import warnings

import sqlalchemy as sql

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode
import pixeltable.exprs as exprs
import pixeltable.catalog as catalog


_logger = logging.getLogger('pixeltable')

class SqlScanNode(ExecNode):
    """Materializes data from the store via SQL
    """
    def __init__(
            self, tbl: catalog.TableVersionPath, row_builder: exprs.RowBuilder,
            select_list: Iterable[exprs.Expr],
            where_clause: Optional[exprs.Expr] = None, filter: Optional[exprs.Predicate] = None,
            order_by_items: Optional[List[Tuple[exprs.Expr, bool]]] = None,
            similarity_clause: Optional[exprs.ImageSimilarityPredicate] = None,
            limit: int = 0, set_pk: bool = False, exact_version_only: Optional[List[catalog.TableVersion]] = None
    ):
        """
        Args:
            select_list: output of the query
            sql_where_clause: SQL Where clause
            filter: additional Where-clause predicate that can't be evaluated via SQL
            limit: max number of rows to return: 0 = no limit
            set_pk: if True, sets the primary for each DataRow
            exact_version_only: tables for which we only want to see rows created at the current version
        """
        # create Select stmt
        if order_by_items is None:
            order_by_items = []
        if exact_version_only is None:
            exact_version_only = []
        super().__init__(row_builder, [], [], None)
        self.tbl = tbl
        target = tbl.tbl_version  # the stored table we're scanning
        self.sql_exprs = exprs.ExprSet(select_list)
        # unstored iter columns: we also need to retrieve whatever is needed to materialize the iter args
        for iter_arg in row_builder.unstored_iter_args.values():
            sql_subexprs = iter_arg.subexprs(filter=lambda e: e.sql_expr() is not None, traverse_matches=False)
            [self.sql_exprs.append(e) for e in sql_subexprs]
        self.filter = filter
        self.filter_eval_ctx = \
            row_builder.create_eval_ctx([filter], exclude=select_list) if filter is not None else None
        self.limit = limit

        # change rowid refs against a base table to rowid refs against the target table, so that we minimize
        # the number of tables that need to be joined to the target table
        for rowid_ref in [e for e in self.sql_exprs if isinstance(e, exprs.RowidRef)]:
            rowid_ref.set_tbl(tbl)

        where_clause_tbl_ids = where_clause.tbl_ids() if where_clause is not None else set()
        refd_tbl_ids = exprs.Expr.list_tbl_ids(self.sql_exprs) | where_clause_tbl_ids
        sql_select_list = [e.sql_expr() for e in self.sql_exprs]
        assert len(sql_select_list) == len(self.sql_exprs)
        assert all([e is not None for e in sql_select_list])
        self.set_pk = set_pk
        self.num_pk_cols = 0
        if set_pk:
            # we also need to retrieve the pk columns
            pk_columns = target.store_tbl.pk_columns()
            self.num_pk_cols = len(pk_columns)
            sql_select_list += pk_columns

        self.stmt = sql.select(*sql_select_list)
        self.stmt = self.create_from_clause(
            tbl, self.stmt, refd_tbl_ids, exact_version_only={t.id for t in exact_version_only})

        # change rowid refs against a base table to rowid refs against the target table, so that we minimize
        # the number of tables that need to be joined to the target table
        for rowid_ref in [e for e, _ in order_by_items if isinstance(e, exprs.RowidRef)]:
            rowid_ref.set_tbl(tbl)
        order_by_clause = [e.sql_expr().desc() if not asc else e.sql_expr() for e, asc in order_by_items]

        if where_clause is not None:
            sql_where_clause = where_clause.sql_expr()
            assert sql_where_clause is not None
            self.stmt = self.stmt.where(sql_where_clause)
        if similarity_clause is not None:
            self.stmt = self.stmt.order_by(
                similarity_clause.img_col_ref.col.sa_idx_col.l2_distance(similarity_clause.embedding()))
        if len(order_by_clause) > 0:
            self.stmt = self.stmt.order_by(*order_by_clause)
        elif target.id in row_builder.unstored_iter_args:
            # we are referencing unstored iter columns from this view and try to order by our primary key,
            # which ensures that iterators will see monotonically increasing pos values
            self.stmt = self.stmt.order_by(*self.tbl.store_tbl.rowid_columns())
        if limit != 0 and self.filter is None:
            # if we need to do post-SQL filtering, we can't use LIMIT
            self.stmt = self.stmt.limit(limit)

        self.result_cursor: Optional[sql.engine.CursorResult] = None

        try:
            # log stmt, if possible
            stmt_str = str(self.stmt.compile(compile_kwargs={'literal_binds': True}))
            _logger.debug(f'SqlScanNode stmt:\n{stmt_str}')
        except Exception as e:
            pass

    @classmethod
    def create_from_clause(
            cls, tbl: catalog.TableVersionPath, stmt: sql.Select, refd_tbl_ids: Optional[Set[UUID]] = None,
            exact_version_only: Optional[Set[UUID]] = None
    ) -> sql.Select:
        """Add From clause to stmt for tables/views referenced by materialized_exprs
        Args:
            tbl: root table of join chain
            stmt: stmt to add From clause to
            materialized_exprs: list of exprs that reference tables in the join chain; if empty, include only the root
            exact_version_only: set of table ids for which we only want to see rows created at the current version
        Returns:
            augmented stmt
        """
        # we need to include at least the root
        if refd_tbl_ids is None:
            refd_tbl_ids = {}
        if exact_version_only is None:
            exact_version_only = {}
        candidates = tbl.get_tbl_versions()
        assert len(candidates) > 0
        joined_tbls: List[catalog.TableVersion] = [candidates[0]]
        for tbl in candidates[1:]:
            if tbl.id in refd_tbl_ids:
                joined_tbls.append(tbl)

        first = True
        for tbl in joined_tbls[::-1]:
            if first:
                stmt = stmt.select_from(tbl.store_tbl.sa_tbl)
                first = False
            else:
                # join tbl to prev_tbl on prev_tbl's rowid cols
                prev_tbl_rowid_cols = prev_tbl.store_tbl.rowid_columns()
                tbl_rowid_cols = tbl.store_tbl.rowid_columns()
                rowid_clauses = \
                    [c1 == c2 for c1, c2 in zip(prev_tbl_rowid_cols, tbl_rowid_cols[:len(prev_tbl_rowid_cols)])]
                stmt = stmt.join(tbl.store_tbl.sa_tbl, sql.and_(*rowid_clauses))
            if tbl.id in exact_version_only:
                stmt = stmt.where(tbl.store_tbl.v_min_col == tbl.version)
            else:
                stmt = stmt \
                    .where(tbl.store_tbl.v_min_col <= tbl.version) \
                    .where(tbl.store_tbl.v_max_col > tbl.version)
            prev_tbl = tbl
        return stmt

    def _log_explain(self, conn: sql.engine.Connection) -> None:
        try:
            # don't set dialect=Env.get().engine.dialect: x % y turns into x %% y, which results in a syntax error
            stmt_str = str(self.stmt.compile(compile_kwargs={'literal_binds': True}))
            explain_result = self.ctx.conn.execute(sql.text(f'EXPLAIN {stmt_str}'))
            explain_str = '\n'.join([str(row) for row in explain_result])
            _logger.debug(f'SqlScanNode explain:\n{explain_str}')
        except Exception as e:
            _logger.warning(f'EXPLAIN failed')

    def __next__(self) -> DataRowBatch:
        if self.result_cursor is None:
            # run the query; do this here rather than in _open(), exceptions are only expected during iteration
            assert self.ctx.conn is not None
            try:
                self._log_explain(self.ctx.conn)
                with warnings.catch_warnings(record=True) as w:
                    self.result_cursor = self.ctx.conn.execute(self.stmt)
                    for warning in w:
                        pass
                self.has_more_rows = True
            except Exception as e:
                self.has_more_rows = False
                raise e

        if not self.has_more_rows:
            raise StopIteration

        output_batch = DataRowBatch(self.tbl.tbl_version, self.row_builder)
        needs_row = True
        while self.ctx.batch_size == 0 or len(output_batch) < self.ctx.batch_size:
            try:
                sql_row = next(self.result_cursor)
            except StopIteration:
                self.has_more_rows = False
                break

            if needs_row:
                output_row = output_batch.add_row()
            if self.num_pk_cols > 0:
                output_row.set_pk(tuple(sql_row[-self.num_pk_cols:]))
            # copy the output of the SQL query into the output row
            for i, e in enumerate(self.sql_exprs):
                slot_idx = e.slot_idx
                output_row[slot_idx] = sql_row[i]
            if self.filter is not None:
                self.row_builder.eval(output_row, self.filter_eval_ctx, profile=self.ctx.profile)
                if output_row[self.filter.slot_idx]:
                    needs_row = True
                    if self.limit is not None and len(output_batch) >= self.limit:
                        self.has_more_rows = False
                        break
                else:
                    # we re-use this row for the next sql row if it didn't pass the filter
                    needs_row = False
                    output_row.clear()

        if not needs_row:
            # the last row didn't pass the filter
            assert self.filter is not None
            output_batch.pop_row()

        _logger.debug(f'SqlScanNode: returning {len(output_batch)} rows')
        if len(output_batch) == 0:
            raise StopIteration
        return output_batch

    def _close(self) -> None:
        if self.result_cursor is not None:
            self.result_cursor.close()




import pytest

import pixeltable as pxt
from pixeltable.iterators import FrameIterator
from pixeltable.tests.utils import get_video_files, skip_test_if_not_installed
from pixeltable.type_system import ImageType, VideoType


class TestNOS:
    def test_basic(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('nos')
        cl = test_client
        video_t = cl.create_table('video_tbl', {'video': VideoType()})
        # create frame view
        args = {'video': video_t.video, 'fps': 1}
        v = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)
        v.add_column(transform1=v.frame.rotate(30), stored=False)
        from pixeltable.functions.nos.object_detection_2d import \
            torchvision_fasterrcnn_mobilenet_v3_large_320_fpn as fasterrcnn
        v.add_column(detections=fasterrcnn(v.transform1))
        from pixeltable.functions.nos.image_embedding import openai_clip
        v.add_column(embed=openai_clip(v.transform1.resize([224, 224])))
        # add a stored column that isn't referenced in nos calls
        v.add_column(transform2=v.frame.rotate(60), stored=True)

        status = video_t.insert(video=get_video_files()[0])
        pass

    def test_exceptions(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('nos')
        cl = test_client
        video_t = cl.create_table('video_tbl', {'video': VideoType()})
        # create frame view
        args = {'video': video_t.video, 'fps': 1}
        v = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)
        video_t.insert(video=get_video_files()[0])

        v.add_column(frame_s=v.frame.resize([640, 480]))
        # 'rotated' has exceptions
        v.add_column(rotated=lambda frame_s, frame_idx: frame_s.rotate(int(360 / frame_idx)), type=ImageType())
        from pixeltable.functions.nos.object_detection_2d import yolox_medium
        v.add_column(detections=yolox_medium(v.rotated), stored=True)
        assert v.where(v.detections.errortype != None).count() == 1

    @pytest.mark.skip(reason='too slow')
    def test_sd(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('nos')
        """Test model that mixes batched with scalar parameters"""
        t = test_client.create_table('sd_test', {'prompt': pxt.StringType()})
        t.insert(prompt='cat on a sofa')
        from pixeltable.functions.nos.image_generation import stabilityai_stable_diffusion_2 as sd2
        t.add_column(img=sd2(t.prompt, 1, 512, 512), stored=True)
        img = t[t.img].show(1)[0, 0]
        assert img.size == (512, 512)



def print_perf_counter_delta(delta: float) -> str:
    """Prints a performance counter delta in a human-readable format.

    Args:
        delta: delta in seconds

    Returns:
        Human-readable string
    """
    if delta < 1e-6:
        return f'{delta * 1e9:.2f} ns'
    elif delta < 1e-3:
        return f'{delta * 1e6:.2f} us'
    elif delta < 1:
        return f'{delta * 1e3:.2f} ms'
    else:
        return f'{delta:.2f} s'



from __future__ import annotations

import inspect
from typing import List, Callable, Optional, overload, Any

import pixeltable as pxt
import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from .batched_function import ExplicitBatchedFunction
from .callable_function import CallableFunction
from .expr_template_function import ExprTemplateFunction
from .function import Function
from .function_registry import FunctionRegistry
from .signature import Signature


# Decorator invoked without parentheses: @pxt.udf
@overload
def udf(decorated_fn: Callable) -> Function: ...


# Decorator schema invoked with parentheses: @pxt.udf(**kwargs)
@overload
def udf(
        *,
        return_type: Optional[ts.ColumnType] = None,
        param_types: Optional[List[ts.ColumnType]] = None,
        batch_size: Optional[int] = None,
        substitute_fn: Optional[Callable] = None,
        _force_stored: bool = False
) -> Callable: ...


def udf(*args, **kwargs):
    """A decorator to create a Function from a function definition.

    Examples:
        >>> @pxt.udf
        ... def my_function(x: int) -> int:
        ...    return x + 1

        >>> @pxt.udf(param_types=[pxt.IntType()], return_type=pxt.IntType())
        ... def my_function(x):
        ...    return x + 1
    """
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):

        # Decorator invoked without parentheses: @pxt.udf
        # Simply call make_function with defaults.
        return make_function(decorated_fn=args[0])

    else:

        # Decorator schema invoked with parentheses: @pxt.udf(**kwargs)
        # Create a decorator for the specified schema.
        return_type = kwargs.pop('return_type', None)
        param_types = kwargs.pop('param_types', None)
        batch_size = kwargs.pop('batch_size', None)
        substitute_fn = kwargs.pop('py_fn', None)
        force_stored = kwargs.pop('_force_stored', False)

        def decorator(decorated_fn: Callable):
            return make_function(
                decorated_fn, return_type, param_types, batch_size, substitute_fn=substitute_fn,
                force_stored=force_stored)

        return decorator


def make_function(
    decorated_fn: Callable,
    return_type: Optional[ts.ColumnType] = None,
    param_types: Optional[List[ts.ColumnType]] = None,
    batch_size: Optional[int] = None,
    substitute_fn: Optional[Callable] = None,
    function_name: Optional[str] = None,
    force_stored: bool = False
) -> Function:
    """
    Constructs a `CallableFunction` or `BatchedFunction`, depending on the
    supplied parameters. If `substitute_fn` is specified, then `decorated_fn`
    will be used only for its signature, with execution delegated to
    `substitute_fn`.
    """
    # Obtain function_path from decorated_fn when appropriate
    if force_stored:
        # force storing the function in the db
        function_path = None
    elif decorated_fn.__module__ != '__main__' and decorated_fn.__name__.isidentifier():
        function_path = f'{decorated_fn.__module__}.{decorated_fn.__qualname__}'
    else:
        function_path = None

    # Derive function_name, if not specified explicitly
    if function_name is None:
        function_name = decorated_fn.__name__

    # Display name to use for error messages
    errmsg_name = function_name if function_path is None else function_path

    sig = Signature.create(decorated_fn, param_types, return_type)

    # batched functions must have a batched return type
    # TODO: remove 'Python' from the error messages when we have full inference with Annotated types
    if batch_size is not None and not sig.is_batched:
        raise excs.Error(f'{errmsg_name}(): batch_size is specified; Python return type must be a `Batch`')
    if batch_size is not None and len(sig.batched_parameters) == 0:
        raise excs.Error(f'{errmsg_name}(): batch_size is specified; at least one Python parameter must be `Batch`')
    if batch_size is None and len(sig.batched_parameters) > 0:
        raise excs.Error(f'{errmsg_name}(): batched parameters in udf, but no `batch_size` given')

    if substitute_fn is None:
        py_fn = decorated_fn
    else:
        if function_path is None:
            raise excs.Error(f'{errmsg_name}(): @udf decorator with a `substitute_fn` can only be used in a module')
        py_fn = substitute_fn

    if batch_size is None:
        result = CallableFunction(signature=sig, py_fn=py_fn, self_path=function_path, self_name=function_name)
    else:
        result = ExplicitBatchedFunction(
            signature=sig, batch_size=batch_size, invoker_fn=py_fn, self_path=function_path)

    # If this function is part of a module, register it
    if function_path is not None:
        FunctionRegistry.get().register_function(function_path, result)

    return result

@overload
def expr_udf(py_fn: Callable) -> ExprTemplateFunction: ...

@overload
def expr_udf(*, param_types: Optional[List[ts.ColumnType]] = None) -> Callable: ...

def expr_udf(*args: Any, **kwargs: Any) -> Any:
    def decorator(py_fn: Callable, param_types: Optional[List[ts.ColumnType]]) -> ExprTemplateFunction:
        if py_fn.__module__ != '__main__' and py_fn.__name__.isidentifier():
            # this is a named function in a module
            function_path = f'{py_fn.__module__}.{py_fn.__qualname__}'
        else:
            function_path = None

        sig = Signature.create(py_fn, param_types=param_types, return_type=None)
        # TODO: verify that the inferred return type matches that of the template
        # TODO: verify that the signature doesn't contain batched parameters

        # construct Parameters from the function signature
        import pixeltable.exprs as exprs
        var_exprs = [exprs.Variable(param.name, param.col_type) for param in sig.parameters.values()]
        # call the function with the parameter expressions to construct an Expr with parameters
        template = py_fn(*var_exprs)
        assert isinstance(template, exprs.Expr)
        py_sig = inspect.signature(py_fn)
        return ExprTemplateFunction(template, py_signature=py_sig, self_path=function_path, name=py_fn.__name__)

    if len(args) == 1:
        assert len(kwargs) == 0 and callable(args[0])
        return decorator(args[0], None)
    else:
        assert len(args) == 0 and len(kwargs) == 1 and 'param_types' in kwargs
        return lambda py_fn: decorator(py_fn, kwargs['param_types'])



import inspect
from typing import List, Dict, Any, Optional, Callable
import abc

from .function import Function
from .signature import Signature


class BatchedFunction(Function):
    """Base class for functions that can run on batches"""

    @abc.abstractmethod
    def get_batch_size(self, *args: Any, **kwargs: Any) -> Optional[int]:
        """Return the batch size for the given arguments, or None if the batch size is unknown.
        args/kwargs might be empty
        """
        raise NotImplementedError

    @abc.abstractmethod
    def invoke(self, arg_batches: List[List[Any]], kwarg_batches: Dict[str, List[Any]]) -> List[Any]:
        """Invoke the function for the given batch and return a batch of results"""
        raise NotImplementedError


class ExplicitBatchedFunction(BatchedFunction):
    """
    A `BatchedFunction` that is defined by a signature and an explicit python
    `Callable`.
    """
    def __init__(self, signature: Signature, batch_size: Optional[int], invoker_fn: Callable, self_path: str):
        super().__init__(signature=signature, py_signature=inspect.signature(invoker_fn), self_path=self_path)
        self.batch_size = batch_size
        self.invoker_fn = invoker_fn

    def get_batch_size(self, *args: Any, **kwargs: Any) -> Optional[int]:
        return self.batch_size

    def invoke(self, arg_batches: List[List[Any]], kwarg_batches: Dict[str, List[Any]]) -> List[Any]:
        """Invoke the function for the given batch and return a batch of results"""
        constant_param_names = [p.name for p in self.signature.constant_parameters]
        kwargs = {k: v[0] for k, v in kwarg_batches.items() if k in constant_param_names}
        kwarg_batches = {k: v for k, v in kwarg_batches.items() if k not in constant_param_names}
        return self.invoker_fn(*arg_batches, **kwargs, **kwarg_batches)

    def validate_call(self, bound_args: Dict[str, Any]) -> None:
        """Verify constant parameters"""
        import pixeltable.exprs as exprs
        for param in self.signature.constant_parameters:
            if param.name in bound_args and isinstance(bound_args[param.name], exprs.Expr):
                raise ValueError(
                    f'{self.display_name}(): '
                    f'parameter {param.name} must be a constant value, not a Pixeltable expression'
                )



from __future__ import annotations

import os
import sys
import warnings
from typing import Optional, Dict, Any, List, Tuple, Set
import logging
import urllib
import sqlalchemy as sql
from tqdm import tqdm, TqdmWarning
import abc

import pixeltable.catalog as catalog
from pixeltable.metadata import schema
from pixeltable.type_system import StringType
from pixeltable.exec import ExecNode
from pixeltable import exprs
from pixeltable.utils.sql import log_stmt, log_explain
import pixeltable.env as env
from pixeltable.utils.media_store import MediaStore


_logger = logging.getLogger('pixeltable')


class StoreBase:
    """Base class for stored tables

    Each row has the following system columns:
    - rowid columns: one or more columns that identify a user-visible row across all versions
    - v_min: version at which the row was created
    - v_max: version at which the row was deleted (or MAX_VERSION if it's still live)
    """

    def __init__(self, tbl_version: catalog.TableVersion):
        self.tbl_version = tbl_version
        self.sa_md = sql.MetaData()
        self.sa_tbl: Optional[sql.Table] = None
        self._create_sa_tbl()

    def pk_columns(self) -> List[sql.Column]:
        return self._pk_columns

    def rowid_columns(self) -> List[sql.Column]:
        return self._pk_columns[:-1]

    @abc.abstractmethod
    def _create_rowid_columns(self) -> List[sql.Column]:
        """Create and return rowid columns"""
        pass

    @abc.abstractmethod
    def _create_system_columns(self) -> List[sql.Column]:
        """Create and return system columns"""
        rowid_cols = self._create_rowid_columns()
        self.v_min_col = sql.Column('v_min', sql.BigInteger, nullable=False)
        self.v_max_col = \
            sql.Column('v_max', sql.BigInteger, nullable=False, server_default=str(schema.Table.MAX_VERSION))
        self._pk_columns = [*rowid_cols, self.v_min_col]
        return [*rowid_cols, self.v_min_col, self.v_max_col]


    def _create_sa_tbl(self) -> None:
        """Create self.sa_tbl from self.tbl_version."""
        system_cols = self._create_system_columns()
        all_cols = system_cols.copy()
        idxs: List[sql.Index] = []
        for col in [c for c in self.tbl_version.cols if c.is_stored]:
            # re-create sql.Column for each column, regardless of whether it already has sa_col set: it was bound
            # to the last sql.Table version we created and cannot be reused
            col.create_sa_cols()
            all_cols.append(col.sa_col)
            if col.records_errors:
                all_cols.append(col.sa_errormsg_col)
                all_cols.append(col.sa_errortype_col)

            if col.is_indexed:
                all_cols.append(col.sa_idx_col)

            # we create an index for:
            # - scalar columns (except for strings, because long strings can't be used for B-tree indices)
            # - non-computed video and image columns (they will contain external paths/urls that users might want to
            #   filter on)
            if (col.col_type.is_scalar_type() and not col.col_type.is_string_type()) \
                    or (col.col_type.is_media_type() and not col.is_computed):
                # index names need to be unique within the Postgres instance
                idx_name = f'idx_{col.id}_{self.tbl_version.id.hex}'
                idxs.append(sql.Index(idx_name, col.sa_col))

        if self.sa_tbl is not None:
            # if we're called in response to a schema change, we need to remove the old table first
            self.sa_md.remove(self.sa_tbl)

        # index for all system columns:
        # - base x view joins can be executed as merge joins
        # - speeds up ORDER BY rowid DESC
        # - allows filtering for a particular table version in index scan
        idx_name = f'sys_cols_idx_{self.tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, *system_cols))
        # v_min/v_max indices: speeds up base table scans needed to propagate a base table insert or delete
        idx_name = f'vmin_idx_{self.tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, self.v_min_col, postgresql_using='brin'))
        idx_name = f'vmax_idx_{self.tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, self.v_max_col, postgresql_using='brin'))

        self.sa_tbl = sql.Table(self._storage_name(), self.sa_md, *all_cols, *idxs)

    @abc.abstractmethod
    def _rowid_join_predicate(self) -> sql.ClauseElement:
        """Return predicate for rowid joins to all bases"""
        pass

    @abc.abstractmethod
    def _storage_name(self) -> str:
        """Return the name of the data store table"""
        pass

    def _move_tmp_media_file(self, file_url: Optional[str], col: catalog.Column, v_min: int) -> str:
        """Move tmp media file with given url to Env.media_dir and return new url, or given url if not a tmp_dir file"""
        pxt_tmp_dir = str(env.Env.get().tmp_dir)
        if file_url is None:
            return None
        parsed = urllib.parse.urlparse(file_url)
        if parsed.scheme != '' and parsed.scheme != 'file':
            # remote url
            return file_url
        file_path = urllib.parse.unquote(parsed.path)
        if not file_path.startswith(pxt_tmp_dir):
            # not a tmp file
            return file_url
        _, ext = os.path.splitext(file_path)
        new_path = str(MediaStore.prepare_media_path(self.tbl_version.id, col.id, v_min, ext=ext))
        os.rename(file_path, new_path)
        new_file_url = urllib.parse.urljoin('file:', urllib.request.pathname2url(new_path))
        return new_file_url

    def _move_tmp_media_files(
            self, table_rows: List[Dict[str, Any]], media_cols: List[catalog.Column], v_min: int
    ) -> None:
        """Move tmp media files that we generated to a permanent location"""
        for c in media_cols:
            for table_row in table_rows:
                file_url = table_row[c.storage_name()]
                table_row[c.storage_name()] = self._move_tmp_media_file(file_url, c, v_min)

    def _create_table_row(
            self, input_row: exprs.DataRow, row_builder: exprs.RowBuilder, media_cols: List[catalog.Column],
            exc_col_ids: Set[int], v_min: int
    ) -> Tuple[Dict[str, Any], int]:
        """Return Tuple[complete table row, # of exceptions] for insert()
        Creates a row that includes the PK columns, with the values from input_row.pk.
        Returns:
            Tuple[complete table row, # of exceptions]
        """
        table_row, num_excs = row_builder.create_table_row(input_row, exc_col_ids)

        assert input_row.pk is not None and len(input_row.pk) == len(self._pk_columns)
        for pk_col, pk_val in zip(self._pk_columns, input_row.pk):
            if pk_col == self.v_min_col:
                table_row[pk_col.name] = v_min
            else:
                table_row[pk_col.name] = pk_val

        return table_row, num_excs

    def count(self) -> None:
        """Return the number of rows visible in self.tbl_version"""
        stmt = sql.select(sql.func.count('*'))\
            .select_from(self.sa_tbl)\
            .where(self.v_min_col <= self.tbl_version.version)\
            .where(self.v_max_col > self.tbl_version.version)
        with env.Env.get().engine.begin() as conn:
            result = conn.execute(stmt).scalar_one()
            assert isinstance(result, int)
            return result

    def create(self, conn: sql.engine.Connection) -> None:
        self.sa_md.create_all(bind=conn)

    def drop(self, conn: sql.engine.Connection) -> None:
        """Drop store table"""
        self.sa_md.drop_all(bind=conn)

    def add_column(self, col: catalog.Column, conn: sql.engine.Connection) -> None:
        """Add column(s) to the store-resident table based on a catalog column

        Note that a computed catalog column will require two extra columns (for the computed value and for the error
        message).
        """
        assert col.is_stored
        stmt = sql.text(f'ALTER TABLE {self._storage_name()} ADD COLUMN {col.storage_name()} {col.col_type.to_sql()}')
        log_stmt(_logger, stmt)
        conn.execute(stmt)
        added_storage_cols = [col.storage_name()]
        if col.records_errors:
            # we also need to create the errormsg and errortype storage cols
            stmt = (f'ALTER TABLE {self._storage_name()} '
                    f'ADD COLUMN {col.errormsg_storage_name()} {StringType().to_sql()} DEFAULT NULL')
            conn.execute(sql.text(stmt))
            stmt = (f'ALTER TABLE {self._storage_name()} '
                    f'ADD COLUMN {col.errortype_storage_name()} {StringType().to_sql()} DEFAULT NULL')
            conn.execute(sql.text(stmt))
        added_storage_cols.extend([col.errormsg_storage_name(), col.errortype_storage_name()])
        self._create_sa_tbl()
        _logger.info(f'Added columns {added_storage_cols} to storage table {self._storage_name()}')

    def drop_column(self, col: Optional[catalog.Column] = None, conn: Optional[sql.engine.Connection] = None) -> None:
        """Re-create self.sa_tbl and drop column, if one is given"""
        if col is not None:
            assert conn is not None
            stmt = f'ALTER TABLE {self._storage_name()} DROP COLUMN {col.storage_name()}'
            conn.execute(sql.text(stmt))
            if col.records_errors:
                stmt = f'ALTER TABLE {self._storage_name()} DROP COLUMN {col.errormsg_storage_name()}'
                conn.execute(sql.text(stmt))
                stmt = f'ALTER TABLE {self._storage_name()} DROP COLUMN {col.errortype_storage_name()}'
                conn.execute(sql.text(stmt))
        self._create_sa_tbl()

    def load_column(
            self, col: catalog.Column, exec_plan: ExecNode, value_expr_slot_idx: int, embedding_slot_idx: int,
            conn: sql.engine.Connection
    ) -> int:
        """Update store column of a computed column with values produced by an execution plan

        Returns:
            number of rows with exceptions
        Raises:
            sql.exc.DBAPIError if there was an error during SQL execution
        """
        num_excs = 0
        num_rows = 0
        for row_batch in exec_plan:
            num_rows += len(row_batch)
            for result_row in row_batch:
                values_dict: Dict[sql.Column, Any] = {}

                if col.is_computed:
                    if result_row.has_exc(value_expr_slot_idx):
                        num_excs += 1
                        value_exc = result_row.get_exc(value_expr_slot_idx)
                        # we store a NULL value and record the exception/exc type
                        error_type = type(value_exc).__name__
                        error_msg = str(value_exc)
                        values_dict = {
                            col.sa_col: None,
                            col.sa_errortype_col: error_type,
                            col.sa_errormsg_col: error_msg
                        }
                    else:
                        val = result_row.get_stored_val(value_expr_slot_idx)
                        if col.col_type.is_media_type():
                            val = self._move_tmp_media_file(val, col, result_row.pk[-1])
                        values_dict = {col.sa_col: val}

                if col.is_indexed:
                    # TODO: deal with exceptions
                    assert not result_row.has_exc(embedding_slot_idx)
                    # don't use get_stored_val() here, we need to pass the ndarray
                    embedding = result_row[embedding_slot_idx]
                    values_dict[col.sa_index_col] = embedding

                update_stmt = sql.update(self.sa_tbl).values(values_dict)
                for pk_col, pk_val in zip(self.pk_columns(), result_row.pk):
                    update_stmt = update_stmt.where(pk_col == pk_val)
                log_stmt(_logger, update_stmt)
                conn.execute(update_stmt)

        return num_excs

    def insert_rows(
            self, exec_plan: ExecNode, conn: sql.engine.Connection, v_min: Optional[int] = None
    ) -> Tuple[int, int, Set[int]]:
        """Insert rows into the store table and update the catalog table's md
        Returns:
            number of inserted rows, number of exceptions, set of column ids that have exceptions
        """
        assert v_min is not None
        exec_plan.ctx.conn = conn
        batch_size = 16  # TODO: is this a good batch size?
        # TODO: total?
        num_excs = 0
        num_rows = 0
        cols_with_excs: Set[int] = set()
        progress_bar: Optional[tqdm] = None  # create this only after we started executing
        row_builder = exec_plan.row_builder
        media_cols = [info.col for info in row_builder.table_columns if info.col.col_type.is_media_type()]
        try:
            exec_plan.open()
            for row_batch in exec_plan:
                num_rows += len(row_batch)
                for batch_start_idx in range(0, len(row_batch), batch_size):
                    # compute batch of rows and convert them into table rows
                    table_rows: List[Dict[str, Any]] = []
                    for row_idx in range(batch_start_idx, min(batch_start_idx + batch_size, len(row_batch))):
                        row = row_batch[row_idx]
                        table_row, num_row_exc = \
                            self._create_table_row(row, row_builder, media_cols, cols_with_excs, v_min=v_min)
                        num_excs += num_row_exc
                        table_rows.append(table_row)
                        if progress_bar is None:
                            warnings.simplefilter("ignore", category=TqdmWarning)
                            progress_bar = tqdm(
                                desc=f'Inserting rows into `{self.tbl_version.name}`',
                                unit=' rows',
                                ncols=100,
                                file=sys.stdout
                            )
                        progress_bar.update(1)
                    self._move_tmp_media_files(table_rows, media_cols, v_min)
                    conn.execute(sql.insert(self.sa_tbl), table_rows)
            if progress_bar is not None:
                progress_bar.close()
            return num_rows, num_excs, cols_with_excs
        finally:
            exec_plan.close()

    def _versions_clause(self, versions: List[Optional[int]], match_on_vmin: bool) -> sql.ClauseElement:
        """Return filter for base versions"""
        v = versions[0]
        if v is None:
            # we're looking at live rows
            clause = sql.and_(self.v_min_col <= self.tbl_version.version, self.v_max_col == schema.Table.MAX_VERSION)
        else:
            # we're looking at a specific version
            clause = self.v_min_col == v if match_on_vmin else self.v_max_col == v
        if len(versions) == 1:
            return clause
        return sql.and_(clause, self.base._versions_clause(versions[1:], match_on_vmin))

    def delete_rows(
            self, current_version: int, base_versions: List[Optional[int]], match_on_vmin: bool,
            where_clause: Optional[sql.ClauseElement], conn: sql.engine.Connection) -> int:
        """Mark rows as deleted that are live and were created prior to current_version.
        Args:
            base_versions: if non-None, join only to base rows that were created at that version,
                otherwise join to rows that are live in the base's current version (which is distinct from the
                current_version parameter)
            match_on_vmin: if True, match exact versions on v_min; if False, match on v_max
            where_clause: if not None, also apply where_clause
        Returns:
            number of deleted rows
        """
        where_clause = sql.true() if where_clause is None else where_clause
        where_clause = sql.and_(
            self.v_min_col < current_version,
            self.v_max_col == schema.Table.MAX_VERSION,
            where_clause)
        rowid_join_clause = self._rowid_join_predicate()
        base_versions_clause = sql.true() if len(base_versions) == 0 \
            else self.base._versions_clause(base_versions, match_on_vmin)
        stmt = sql.update(self.sa_tbl) \
            .values({self.v_max_col: current_version}) \
            .where(where_clause) \
            .where(rowid_join_clause) \
            .where(base_versions_clause)
        log_explain(_logger, stmt, conn)
        status = conn.execute(stmt)
        return status.rowcount


class StoreTable(StoreBase):
    def __init__(self, tbl_version: catalog.TableVersion):
        assert not tbl_version.is_view()
        super().__init__(tbl_version)

    def _create_rowid_columns(self) -> List[sql.Column]:
        self.rowid_col = sql.Column('rowid', sql.BigInteger, nullable=False)
        return [self.rowid_col]

    def _storage_name(self) -> str:
        return f'tbl_{self.tbl_version.id.hex}'

    def _rowid_join_predicate(self) -> sql.ClauseElement:
        return sql.true()


class StoreView(StoreBase):
    def __init__(self, catalog_view: catalog.TableVersion):
        assert catalog_view.is_view()
        self.base = catalog_view.base.store_tbl
        super().__init__(catalog_view)

    def _create_rowid_columns(self) -> List[sql.Column]:
        # a view row corresponds directly to a single base row, which means it needs to duplicate its rowid columns
        self.rowid_cols = [sql.Column(c.name, c.type) for c in self.base.rowid_columns()]
        return self.rowid_cols

    def _storage_name(self) -> str:
        return f'view_{self.tbl_version.id.hex}'

    def _rowid_join_predicate(self) -> sql.ClauseElement:
        return sql.and_(
            self.base._rowid_join_predicate(),
            *[c1 == c2 for c1, c2 in zip(self.rowid_columns(), self.base.rowid_columns())])

class StoreComponentView(StoreView):
    """A view that stores components of its base, as produced by a ComponentIterator

    PK: now also includes pos, the position returned by the ComponentIterator for the base row identified by base_rowid
    """
    def __init__(self, catalog_view: catalog.TableVersion):
        super().__init__(catalog_view)

    def _create_rowid_columns(self) -> List[sql.Column]:
        # each base row is expanded into n view rows
        self.rowid_cols = [sql.Column(c.name, c.type) for c in self.base.rowid_columns()]
        # name of pos column: avoid collisions with bases' pos columns
        self.pos_col = sql.Column(f'pos_{len(self.rowid_cols) - 1}', sql.BigInteger, nullable=False)
        self.pos_col_idx = len(self.rowid_cols)
        self.rowid_cols.append(self.pos_col)
        return self.rowid_cols

    def _create_sa_tbl(self) -> None:
        super()._create_sa_tbl()
        # we need to fix up the 'pos' column in TableVersion
        self.tbl_version.cols_by_name['pos'].sa_col = self.pos_col

    def _rowid_join_predicate(self) -> sql.ClauseElement:
        return sql.and_(
            self.base._rowid_join_predicate(),
            *[c1 == c2 for c1, c2 in zip(self.rowid_columns()[:-1], self.base.rowid_columns())])



import inspect
from typing import Dict, Optional, Callable, List

import pixeltable
import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from .function import Function
from .signature import Signature, Parameter


class ExprTemplateFunction(Function):
    """A parameterized expression from which an executable Expr is created with a function call."""

    def __init__(
            self, expr: 'pixeltable.exprs.Expr', py_signature: inspect.Signature, self_path: Optional[str] = None,
            name: Optional[str] = None):
        import pixeltable.exprs as exprs
        self.expr = expr
        self.self_name = name
        self.param_exprs = list(set(expr.subexprs(expr_class=exprs.Variable)))
        # make sure there are no duplicate names
        assert len(self.param_exprs) == len(set(p.name for p in self.param_exprs))
        self.param_exprs_by_name = {p.name: p for p in self.param_exprs}

        # verify default values
        self.defaults: Dict[str, exprs.Literal] = {}  # key: param name, value: default value converted to a Literal
        for py_param in py_signature.parameters.values():
            if py_param.default is inspect.Parameter.empty:
                continue
            param_expr = self.param_exprs_by_name[py_param.name]
            try:
                literal_default = exprs.Literal(py_param.default, col_type=param_expr.col_type)
                self.defaults[py_param.name] = literal_default
            except TypeError as e:
                msg = str(e)
                raise excs.Error(f"Default value for parameter '{py_param.name}': {msg[0].lower() + msg[1:]}")
        # construct signature
        assert len(self.param_exprs) == len(py_signature.parameters)
        fn_params = [
            Parameter(p.name, self.param_exprs_by_name[p.name].col_type, p.kind)
            for p in py_signature.parameters.values()
        ]
        signature = Signature(return_type=expr.col_type, parameters=fn_params)

        super().__init__(signature, py_signature=py_signature, self_path=self_path)

    def instantiate(self, *args: object, **kwargs: object) -> 'pixeltable.exprs.Expr':
        bound_args = self.py_signature.bind(*args, **kwargs).arguments
        # apply defaults, otherwise we might have Parameters left over
        bound_args.update(
            {param_name: default for param_name, default in self.defaults.items() if param_name not in bound_args})
        result = self.expr.copy()
        for param_name, arg in bound_args.items():
            param_expr = self.param_exprs_by_name[param_name]
            result = result.substitute(param_expr, arg)
        import pixeltable.exprs as exprs
        assert not result.contains(exprs.Variable)
        return result

    @property
    def display_name(self) -> str:
        return self.self_name

    @property
    def name(self) -> str:
        return self.self_name

    def _as_dict(self) -> Dict:
        if self.self_path is not None:
            return super()._as_dict()
        return {
            'name': self.name,
            'expr': self.expr.as_dict(),
            **super()._as_dict()
        }

    @classmethod
    def _from_dict(cls, d: Dict) -> Function:
        if 'expr' not in d:
            return super()._from_dict(d)
        import pixeltable.exprs as exprs
        return cls(exprs.Expr.from_dict(d['expr']), name=d['name'])



from __future__ import annotations
import logging
from typing import List, Optional, Type, Dict, Set, Any
from uuid import UUID
import inspect

import sqlalchemy.orm as orm

from .table import Table
from .table_version import TableVersion
from .table_version_path import TableVersionPath
from .column import Column
from .catalog import Catalog
from .globals import POS_COLUMN_NAME
from pixeltable.env import Env
from pixeltable.iterators import ComponentIterator
from pixeltable.exceptions import Error
import pixeltable.func as func
import pixeltable.type_system as ts
import pixeltable.catalog as catalog
import pixeltable.metadata.schema as md_schema
from pixeltable.type_system import InvalidType, IntType
import pixeltable.exceptions as excs


_logger = logging.getLogger('pixeltable')

class View(Table):
    """A `Table` that presents a virtual view of another table (or view).

    A view is typically backed by a store table, which records the view's columns and is joined back to the bases
    at query execution time.
    The exception is a snapshot view without a predicate and without additional columns: in that case, the view
    is simply a reference to a specific set of base versions.
    """
    def __init__(
            self, id: UUID, dir_id: UUID, name: str, tbl_version_path: TableVersionPath, base: Table,
            snapshot_only: bool):
        super().__init__(id, dir_id, name, tbl_version_path)
        self._base = base  # keep a reference to the base Table, so that we can keep track of its dependents
        self._snapshot_only = snapshot_only

    @classmethod
    def display_name(cls) -> str:
        return 'view'

    @classmethod
    def create(
            cls, dir_id: UUID, name: str, base: Table, schema: Dict[str, Any],
            predicate: 'exprs.Predicate', is_snapshot: bool, num_retained_versions: int, comment: str,
            iterator_cls: Optional[Type[ComponentIterator]], iterator_args: Optional[Dict]
    ) -> View:
        columns = cls._create_columns(schema)
        cls._verify_schema(columns)

        # verify that filter can be evaluated in the context of the base
        if predicate is not None:
            if not predicate.is_bound_by(base.tbl_version_path):
                raise excs.Error(f'Filter cannot be computed in the context of the base {base._name}')
            # create a copy that we can modify and store
            predicate = predicate.copy()

        # same for value exprs
        for col in columns:
            if not col.is_computed:
                continue
            # make sure that the value can be computed in the context of the base
            if col.value_expr is not None and not col.value_expr.is_bound_by(base.tbl_version_path):
                raise excs.Error(
                    f'Column {col.name}: value expression cannot be computed in the context of the base {base._name}')

        if iterator_cls is not None:
            assert iterator_args is not None

            # validate iterator_args
            py_signature = inspect.signature(iterator_cls.__init__)
            try:
                # make sure iterator_args can be used to instantiate iterator_cls
                bound_args = py_signature.bind(None, **iterator_args).arguments  # None: arg for self
                # we ignore 'self'
                first_param_name = next(iter(py_signature.parameters))  # can't guarantee it's actually 'self'
                del bound_args[first_param_name]

                # construct Signature and type-check bound_args
                params = [
                    func.Parameter(param_name, param_type, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                    for param_name, param_type in iterator_cls.input_schema().items()
                ]
                sig = func.Signature(InvalidType(), params)
                from pixeltable.exprs import FunctionCall
                FunctionCall.check_args(sig, bound_args)
            except TypeError as e:
                raise Error(f'Cannot instantiate iterator with given arguments: {e}')

            # prepend pos and output_schema columns to cols:
            # a component view exposes the pos column of its rowid;
            # we create that column here, so it gets assigned a column id;
            # stored=False: it is not stored separately (it's already stored as part of the rowid)
            iterator_cols = [Column(POS_COLUMN_NAME, IntType(), stored=False)]
            output_dict, unstored_cols = iterator_cls.output_schema(**bound_args)
            iterator_cols.extend([
                Column(col_name, col_type, stored=col_name not in unstored_cols)
                for col_name, col_type in output_dict.items()
            ])

            iterator_col_names = {col.name for col in iterator_cols}
            for col in columns:
                if col.name in iterator_col_names:
                    raise Error(f'Duplicate name: column {col.name} is already present in the iterator output schema')
            columns = iterator_cols + columns

        with orm.Session(Env.get().engine, future=True) as session:
            from pixeltable.exprs import InlineDict
            iterator_args_expr = InlineDict(iterator_args) if iterator_args is not None else None
            iterator_class_fqn = f'{iterator_cls.__module__}.{iterator_cls.__name__}' if iterator_cls is not None \
                else None
            base_version_path = cls._get_snapshot_path(base.tbl_version_path) if is_snapshot else base.tbl_version_path
            base_versions = [
                (tbl_version.id.hex, tbl_version.version if is_snapshot or tbl_version.is_snapshot else None)
                for tbl_version in base_version_path.get_tbl_versions()
            ]

            # if this is a snapshot, we need to retarget all exprs to the snapshot tbl versions
            if is_snapshot:
                predicate = predicate.retarget(base_version_path) if predicate is not None else None
                iterator_args_expr = iterator_args_expr.retarget(base_version_path) \
                    if iterator_args_expr is not None else None
                for col in columns:
                    if col.value_expr is not None:
                        col.value_expr = col.value_expr.retarget(base_version_path)

            view_md = md_schema.ViewMd(
                is_snapshot=is_snapshot, predicate=predicate.as_dict() if predicate is not None else None,
                base_versions=base_versions,
                iterator_class_fqn=iterator_class_fqn,
                iterator_args=iterator_args_expr.as_dict() if iterator_args_expr is not None else None)

            id, tbl_version = TableVersion.create(
                session, dir_id, name, columns, num_retained_versions, comment, base_path=base_version_path, view_md=view_md)
            if tbl_version is None:
                # this is purely a snapshot: we use the base's tbl version path
                view = cls(id, dir_id, name, base_version_path, base, snapshot_only=True)
                _logger.info(f'created snapshot {name}')
            else:
                view = cls(
                    id, dir_id, name, TableVersionPath(tbl_version, base=base_version_path), base,
                    snapshot_only=False)
                _logger.info(f'Created view `{name}`, id={tbl_version.id}')

                from pixeltable.plan import Planner
                plan, num_values_per_row = Planner.create_view_load_plan(view.tbl_version_path)
                num_rows, num_excs, cols_with_excs = tbl_version.store_tbl.insert_rows(
                    plan, session.connection(), v_min=tbl_version.version)
                print(f'Created view `{name}` with {num_rows} rows, {num_excs} exceptions.')

            session.commit()
            cat = Catalog.get()
            cat.tbl_dependents[view._id] = []
            cat.tbl_dependents[base._id].append(view)
            cat.tbls[view._id] = view
            return view

    @classmethod
    def _verify_column(cls, col: Column, existing_column_names: Set[str]) -> None:
        # make sure that columns are nullable or have a default
        if not col.col_type.nullable and not col.is_computed:
            raise Error(f'Column {col.name}: non-computed columns in views must be nullable')
        super()._verify_column(col, existing_column_names)

    @classmethod
    def _get_snapshot_path(cls, tbl_version_path: TableVersionPath) -> TableVersionPath:
        """Returns snapshot of the given table version path.
        All TableVersions of that path will be snapshot versions. Creates new versions from mutable versions,
        if necessary.
        """
        if tbl_version_path.is_snapshot():
            return tbl_version_path
        tbl_version = tbl_version_path.tbl_version
        if not tbl_version.is_snapshot:
            # create and register snapshot version
            tbl_version = tbl_version.create_snapshot_copy()
            assert tbl_version.is_snapshot

        return TableVersionPath(
            tbl_version,
            base=cls._get_snapshot_path(tbl_version_path.base) if tbl_version_path.base is not None else None)

    def _drop(self) -> None:
        cat = catalog.Catalog.get()
        if self._snapshot_only:
            # there is not TableVersion to drop
            self._check_is_dropped()
            self.is_dropped = True
            with Env.get().engine.begin() as conn:
                TableVersion.delete_md(self._id, conn)
            # update catalog
            cat = catalog.Catalog.get()
            del cat.tbls[self._id]
        else:
            super()._drop()
        cat.tbl_dependents[self._base._id].remove(self)
        del cat.tbl_dependents[self._id]




from __future__ import annotations
from typing import List, Tuple, Dict
from collections import defaultdict
import sys

import numpy as np

import pixeltable.type_system as ts
import pixeltable.func as func


# TODO: figure out a better submodule structure

# the following function has been adapted from MMEval
# (sources at https://github.com/open-mmlab/mmeval)
# Copyright (c) OpenMMLab. All rights reserved.
def calculate_bboxes_area(bboxes: np.ndarray) -> np.ndarray:
    """Calculate area of bounding boxes.

    Args:
        bboxes (numpy.ndarray): The bboxes with shape (n, 4) or (4, ) in 'xyxy' format.
     Returns:
        numpy.ndarray: The area of bboxes.
    """
    bboxes_w = (bboxes[..., 2] - bboxes[..., 0])
    bboxes_h = (bboxes[..., 3] - bboxes[..., 1])
    areas = bboxes_w * bboxes_h
    return areas

# the following function has been adapted from MMEval
# (sources at https://github.com/open-mmlab/mmeval)
# Copyright (c) OpenMMLab. All rights reserved.
def calculate_overlaps(bboxes1: np.ndarray, bboxes2: np.ndarray) -> np.ndarray:
    """Calculate the overlap between each bbox of bboxes1 and bboxes2.

    Args:
        bboxes1 (numpy.ndarray): The bboxes with shape (n, 4) in 'xyxy' format.
        bboxes2 (numpy.ndarray): The bboxes with shape (k, 4) in 'xyxy' format.
    Returns:
        numpy.ndarray: IoUs or IoFs with shape (n, k).
    """
    bboxes1 = bboxes1.astype(np.float32)
    bboxes2 = bboxes2.astype(np.float32)
    rows = bboxes1.shape[0]
    cols = bboxes2.shape[0]
    overlaps = np.zeros((rows, cols), dtype=np.float32)

    if rows * cols == 0:
        return overlaps

    if bboxes1.shape[0] > bboxes2.shape[0]:
        # Swap bboxes for faster calculation.
        bboxes1, bboxes2 = bboxes2, bboxes1
        overlaps = np.zeros((cols, rows), dtype=np.float32)
        exchange = True
    else:
        exchange = False

    # Calculate the bboxes area.
    area1 = calculate_bboxes_area(bboxes1)
    area2 = calculate_bboxes_area(bboxes2)
    eps = np.finfo(np.float32).eps

    for i in range(bboxes1.shape[0]):
        x_start = np.maximum(bboxes1[i, 0], bboxes2[:, 0])
        y_start = np.maximum(bboxes1[i, 1], bboxes2[:, 1])
        x_end = np.minimum(bboxes1[i, 2], bboxes2[:, 2])
        y_end = np.minimum(bboxes1[i, 3], bboxes2[:, 3])
        overlap_w = np.maximum(x_end - x_start, 0)
        overlap_h = np.maximum(y_end - y_start, 0)
        overlap = overlap_w * overlap_h

        union = area1[i] + area2 - overlap
        union = np.maximum(union, eps)
        overlaps[i, :] = overlap / union
    return overlaps if not exchange else overlaps.T


# the following function has been adapted from MMEval
# (sources at https://github.com/open-mmlab/mmeval)
# Copyright (c) OpenMMLab. All rights reserved.
def calculate_image_tpfp(
    pred_bboxes: np.ndarray, pred_scores: np.ndarray, gt_bboxes: np.ndarray, min_iou: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate the true positive and false positive on an image.

    Args:
        pred_bboxes (numpy.ndarray): Predicted bboxes of this image, with
            shape (N, 5). The scores The predicted score of the bbox is
            concatenated behind the predicted bbox.
        gt_bboxes (numpy.ndarray): Ground truth bboxes of this image, with
            shape (M, 4).
        min_iou (float): The IoU threshold.

    Returns:
        tuple (tp, fp):

        - tp (numpy.ndarray): Shape (N,),
          the true positive flag of each predicted bbox on this image.
        - fp (numpy.ndarray): Shape (N,),
          the false positive flag of each predicted bbox on this image.
    """
    # Step 1. Concatenate `gt_bboxes` and `ignore_gt_bboxes`, then set
    # the `ignore_gt_flags`.
    # all_gt_bboxes = np.concatenate((gt_bboxes, ignore_gt_bboxes))
    # ignore_gt_flags = np.concatenate((np.zeros(
    #     (gt_bboxes.shape[0], 1),
    #     dtype=bool), np.ones((ignore_gt_bboxes.shape[0], 1), dtype=bool)))

    # Step 2. Initialize the `tp` and `fp` arrays.
    num_preds = pred_bboxes.shape[0]
    tp = np.zeros(num_preds, dtype=np.int8)
    fp = np.zeros(num_preds, dtype=np.int8)

    # Step 3. If there are no gt bboxes in this image, then all pred bboxes
    # within area range are false positives.
    if gt_bboxes.shape[0] == 0:
        fp[...] = 1
        return tp, fp

    # Step 4. Calculate the IoUs between the predicted bboxes and the
    # ground truth bboxes.
    ious = calculate_overlaps(pred_bboxes, gt_bboxes)
    # For each pred bbox, the max iou with all gts.
    ious_max = ious.max(axis=1)
    # For each pred bbox, which gt overlaps most with it.
    ious_argmax = ious.argmax(axis=1)
    # Sort all pred bbox in descending order by scores.
    sorted_indices = np.argsort(-pred_scores)

    # Step 5. Count the `tp` and `fp` of each iou threshold and area range.
    # The flags that gt bboxes have been matched.
    gt_covered_flags = np.zeros(gt_bboxes.shape[0], dtype=bool)

    # Count the prediction bboxes in order of decreasing score.
    for pred_bbox_idx in sorted_indices:
        if ious_max[pred_bbox_idx] >= min_iou:
            matched_gt_idx = ious_argmax[pred_bbox_idx]
            if not gt_covered_flags[matched_gt_idx]:
                tp[pred_bbox_idx] = 1
                gt_covered_flags[matched_gt_idx] = True
            else:
                # This gt bbox has been matched and counted as fp.
                fp[pred_bbox_idx] = 1
        else:
            fp[pred_bbox_idx] = 1

    return tp, fp

@func.udf(
    return_type=ts.JsonType(nullable=False),
    param_types=[
        ts.JsonType(nullable=False),
        ts.JsonType(nullable=False),
        ts.JsonType(nullable=False),
        ts.JsonType(nullable=False),
        ts.JsonType(nullable=False)
    ])
def eval_detections(
        pred_bboxes: List[List[int]], pred_classes: List[int], pred_scores: List[float],
        gt_bboxes: List[List[int]], gt_classes: List[int]
) -> Dict:
    class_idxs = list(set(pred_classes + gt_classes))
    result: List[Dict] = []
    pred_bboxes_arr = np.asarray(pred_bboxes)
    pred_classes_arr = np.asarray(pred_classes)
    pred_scores_arr = np.asarray(pred_scores)
    gt_bboxes_arr = np.asarray(gt_bboxes)
    gt_classes_arr = np.asarray(gt_classes)
    for class_idx in class_idxs:
        pred_filter = pred_classes_arr == class_idx
        gt_filter = gt_classes_arr == class_idx
        class_pred_scores = pred_scores_arr[pred_filter]
        tp, fp = calculate_image_tpfp(
            pred_bboxes_arr[pred_filter], class_pred_scores, gt_bboxes_arr[gt_filter], [0.5])
        ordered_class_pred_scores = -np.sort(-class_pred_scores)
        result.append({
            'min_iou': 0.5, 'class': class_idx, 'tp': tp.tolist(), 'fp': fp.tolist(),
            'scores': ordered_class_pred_scores.tolist(), 'num_gts': gt_filter.sum().item(),
        })
    return result

@func.uda(
    update_types=[ts.JsonType()], value_type=ts.JsonType(), name='mean_ap', allows_std_agg=True, allows_window=False)
class MeanAPAggregator:
    def __init__(self):
        self.class_tpfp: Dict[int, List[Dict]] = defaultdict(list)

    def update(self, eval_dicts: List[Dict]) -> None:
        for eval_dict in eval_dicts:
            class_idx = eval_dict['class']
            self.class_tpfp[class_idx].append(eval_dict)

    def value(self) -> Dict:
        eps = np.finfo(np.float32).eps
        result: Dict[int, float] = {}
        for class_idx, tpfp in self.class_tpfp.items():
            a1 = [x['tp'] for x in tpfp]
            tp = np.concatenate([x['tp'] for x in tpfp], axis=0)
            fp = np.concatenate([x['fp'] for x in tpfp], axis=0)
            num_gts = np.sum([x['num_gts'] for x in tpfp])
            scores = np.concatenate([np.asarray(x['scores']) for x in tpfp])
            sorted_idxs = np.argsort(-scores)
            tp_cumsum = tp[sorted_idxs].cumsum()
            fp_cumsum = fp[sorted_idxs].cumsum()
            precision = tp_cumsum / np.maximum(tp_cumsum + fp_cumsum, eps)
            recall = tp_cumsum / np.maximum(num_gts, eps)

            mrec = np.hstack((0, recall, 1))
            mpre = np.hstack((0, precision, 0))
            for i in range(mpre.shape[0] - 1, 0, -1):
                mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])
            ind = np.where(mrec[1:] != mrec[:-1])[0]
            ap = np.sum((mrec[ind + 1] - mrec[ind]) * mpre[ind + 1])
            result[class_idx] = ap.item()
        return result



from typing import List, Dict, Any, Optional
import urllib
import logging
import os

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode
import pixeltable.catalog as catalog
import pixeltable.exprs as exprs
import pixeltable.env as env
from pixeltable.utils.media_store import MediaStore


_logger = logging.getLogger('pixeltable')

class InMemoryDataNode(ExecNode):
    """Outputs in-memory data as a row batch of a particular table"""
    def __init__(
            self, tbl: catalog.TableVersionPath, rows: List[Dict[str, Any]],
            row_builder: exprs.RowBuilder, start_row_id: int,
    ):
        super().__init__(row_builder, [], [], None)
        assert tbl.is_insertable()
        self.tbl = tbl
        self.input_rows = rows
        self.start_row_id = start_row_id
        self.has_returned_data = False
        self.output_rows: Optional[DataRowBatch] = None

    def _open(self) -> None:
        """Create row batch and populate with self.input_rows"""
        column_info = {info.col.name: info for info in self.row_builder.output_slot_idxs()}
        # stored columns that are not computed
        inserted_column_names = set([
            info.col.name for info in self.row_builder.output_slot_idxs()
            if info.col.is_stored and not info.col.is_computed
        ])

        self.output_rows = DataRowBatch(self.tbl, self.row_builder, len(self.input_rows))
        for row_idx, input_row in enumerate(self.input_rows):
            # populate the output row with the values provided in the input row
            for col_name, val in input_row.items():
                col_info = column_info.get(col_name)
                assert col_info is not None

                if col_info.col.col_type.is_image_type() and isinstance(val, bytes):
                    # this is a literal image, ie, a sequence of bytes; we save this as a media file and store the path
                    path = str(MediaStore.prepare_media_path(self.tbl.id, col_info.col.id, self.tbl.version))
                    open(path, 'wb').write(val)
                    val = path
                self.output_rows[row_idx][col_info.slot_idx] = val

            # set the remaining stored non-computed columns to null
            null_col_names = inserted_column_names - set(input_row.keys())
            for col_name in null_col_names:
                col_info = column_info.get(col_name)
                assert col_info is not None
                self.output_rows[row_idx][col_info.slot_idx] = None

        self.output_rows.set_row_ids([self.start_row_id + i for i in range(len(self.output_rows))])
        self.ctx.num_rows = len(self.output_rows)

    def __next__(self) -> DataRowBatch:
        if self.has_returned_data:
            raise StopIteration
        self.has_returned_data = True
        _logger.debug(f'InMemoryDataNode: created row batch with {len(self.output_rows)} output_rows')
        return self.output_rows




from typing import Optional

import numpy as np
import pytest

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable import catalog
from pixeltable.func import Function, FunctionRegistry, Batch
from pixeltable.type_system import IntType, FloatType
from pixeltable.tests.utils import assert_resultset_eq
import pixeltable.func as func


def dummy_fn(i: int) -> int:
    return i

class TestFunction:
    @pxt.udf(return_type=IntType(), param_types=[IntType()])
    def func(x: int) -> int:
        return x + 1

    @pxt.uda(name='agg', value_type=IntType(), update_types=[IntType()])
    class Aggregator:
        def __init__(self):
            self.sum = 0
        def update(self, val: int) -> None:
            if val is not None:
                self.sum += val
        def value(self) -> int:
            return self.sum

    def test_serialize_anonymous(self, init_env) -> None:
        d = self.func.as_dict()
        FunctionRegistry.get().clear_cache()
        deserialized = Function.from_dict(d)
        # TODO: add Function.exec() and then use that
        assert deserialized.py_fn(1) == 2

    @pytest.mark.skip(reason='deprecated')
    def test_create(self, test_client: pxt.Client) -> None:
        cl = test_client
        cl.create_function('test_fn', self.func)
        assert self.func.md.fqn == 'test_fn'
        FunctionRegistry.get().clear_cache()
        cl = pxt.Client(reload=True)
        _ = cl.list_functions()
        fn2 = cl.get_function('test_fn')
        assert fn2.md.fqn == 'test_fn'
        assert fn2.py_fn(1) == 2

        with pytest.raises(excs.Error):
            cl.create_function('test_fn', self.func)
        with pytest.raises(excs.Error):
            cl.create_function('dir1.test_fn', self.func)
        with pytest.raises(excs.Error):
            library_fn = make_library_function(IntType(), [IntType()], __name__, 'dummy_fn')
            cl.create_function('library_fn', library_fn)

    @pytest.mark.skip(reason='deprecated')
    def test_update(self, test_client: pxt.Client, test_tbl: catalog.Table) -> None:
        cl = test_client
        t = test_tbl
        cl.create_function('test_fn', self.func)
        res1 = t[self.func(t.c2)].show(0).to_pandas()

        # load function from db and make sure it computes the same thing as before
        FunctionRegistry.get().clear_cache()
        cl = pxt.Client(reload=True)
        fn = cl.get_function('test_fn')
        res2 = t[fn(t.c2)].show(0).to_pandas()
        assert res1.col_0.equals(res2.col_0)
        fn.py_fn = lambda x: x + 2
        cl.update_function('test_fn', fn)
        assert self.func.md.fqn == fn.md.fqn  # fqn doesn't change

        FunctionRegistry.get().clear_cache()
        cl = pxt.Client(reload=True)
        fn = cl.get_function('test_fn')
        assert self.func.md.fqn == fn.md.fqn  # fqn doesn't change
        res3 = t[fn(t.c2)].show(0).to_pandas()
        assert (res2.col_0 + 1).equals(res3.col_0)

        # signature changes
        with pytest.raises(excs.Error):
            cl.update_function('test_fn', make_function(FloatType(), [IntType()], fn.py_fn))
        with pytest.raises(excs.Error):
            cl.update_function('test_fn', make_function(IntType(), [FloatType()], fn.py_fn))
        with pytest.raises(excs.Error):
            cl.update_function('test_fn', self.agg)

    @pytest.mark.skip(reason='deprecated')
    def test_move(self, test_client: pxt.Client) -> None:
        cl = test_client
        cl.create_function('test_fn', self.func)

        FunctionRegistry.get().clear_cache()
        cl = pxt.Client(reload=True)
        with pytest.raises(excs.Error):
            cl.move('test_fn2', 'test_fn')
        cl.move('test_fn', 'test_fn2')
        func = cl.get_function('test_fn2')
        assert func.py_fn(1) == 2
        assert func.md.fqn == 'test_fn2'

        with pytest.raises(excs.Error):
            _ = cl.get_function('test_fn')

        # move function between directories
        cl.create_dir('functions')
        cl.create_dir('functions2')
        cl.create_function('functions.func1', self.func)
        with pytest.raises(excs.Error):
            cl.move('functions2.func1', 'functions.func1')
        cl.move('functions.func1', 'functions2.func1')
        func = cl.get_function('functions2.func1')
        assert func.md.fqn == 'functions2.func1'


        FunctionRegistry.get().clear_cache()
        cl = pxt.Client(reload=True)
        func = cl.get_function('functions2.func1')
        assert func.py_fn(1) == 2
        assert func.md.fqn == 'functions2.func1'
        with pytest.raises(excs.Error):
            _ = cl.get_function('functions.func1')

    @pytest.mark.skip(reason='deprecated')
    def test_drop(self, test_client: pxt.Client) -> None:
        cl = test_client
        cl.create_function('test_fn', self.func)
        FunctionRegistry.get().clear_cache()
        cl = pxt.Client(reload=True)
        cl.drop_function('test_fn')

        with pytest.raises(excs.Error):
            _ = cl.get_function('test_fn')

    def test_list(self, test_client: pxt.Client) -> None:
        _ = FunctionRegistry.get().list_functions()
        print(_)

    def test_stored_udf(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = cl.create_table('test', {'c1': pxt.IntType(), 'c2': pxt.FloatType()})
        rows = [{'c1': i, 'c2': i + 0.5} for i in range(100)]
        status = t.insert(rows)
        assert status.num_rows == len(rows)
        assert status.num_excs == 0

        @pxt.udf(_force_stored=True)
        def f1(a: int, b: float) -> float:
            return a + b
        t['f1'] = f1(t.c1, t.c2)

        func.FunctionRegistry.get().clear_cache()
        cl = pxt.Client(reload=True)
        t = cl.get_table('test')
        status = t.insert(rows)
        assert status.num_rows == len(rows)
        assert status.num_excs == 0

    def test_call(self, test_tbl: catalog.Table) -> None:
        t = test_tbl

        @pxt.udf(return_type=IntType(), param_types=[IntType(), FloatType(), FloatType(), FloatType()])
        def f1(a: int, b: float, c: float = 0.0, d: float = 1.0) -> float:
            return a + b + c + d

        r0 = t[t.c2, t.c3].show(0).to_pandas()
        # positional params with default args
        r1 = t[f1(t.c2, t.c3)].show(0).to_pandas()['col_0']
        assert np.all(r1 == r0.c2 + r0.c3 + 1.0)
        # kw args only
        r2 = t[f1(c=0.0, b=t.c3, a=t.c2)].show(0).to_pandas()['col_0']
        assert np.all(r1 == r2)
        # overriding default args
        r3 = t[f1(d=0.0, c=1.0, b=t.c3, a=t.c2)].show(0).to_pandas()['col_0']
        assert np.all(r2 == r3)
        # overriding default with positional arg
        r4 = t[f1(t.c2, t.c3, 0.0)].show(0).to_pandas()['col_0']
        assert np.all(r3 == r4)
        # overriding default with positional arg and kw arg
        r5 = t[f1(t.c2, t.c3, 1.0, d=0.0)].show(0).to_pandas()['col_0']
        assert np.all(r4 == r5)
        # d is kwarg
        r6 = t[f1(t.c2, d=1.0, b=t.c3)].show(0).to_pandas()['col_0']
        assert np.all(r5 == r6)
        # d is Expr kwarg
        r6 = t[f1(1, d=t.c3, b=t.c3)].show(0).to_pandas()['col_0']
        assert np.all(r5 == r6)

        # test handling of Nones
        @pxt.udf(
            return_type=IntType(),
            param_types=[IntType(nullable=True), FloatType(nullable=False), FloatType(nullable=True)])
        def f2(a: int, b: float = 0.0, c: float = 1.0) -> float:
            return (0.0 if a is None else a) + b + (0.0 if c is None else c)
        r0 = t[f2(1, t.c3)].show(0).to_pandas()['col_0']
        r1 = t[f2(None, t.c3, 2.0)].show(0).to_pandas()['col_0']
        assert np.all(r0 == r1)
        r2 = t[f2(2, t.c3, None)].show(0).to_pandas()['col_0']
        assert np.all(r1 == r2)
        # kwarg with None
        r3 = t[f2(c=None, a=t.c2)].show(0).to_pandas()['col_0']
        # kwarg with Expr
        r4 = t[f2(c=t.c3, a=None)].show(0).to_pandas()['col_0']
        assert np.all(r3 == r4)

        with pytest.raises(TypeError) as exc_info:
            _ = t[f1(t.c2, c=0.0)].show(0)
        assert "'b'" in str(exc_info.value)
        with pytest.raises(TypeError) as exc_info:
            _ = t[f1(t.c2)].show(0)
        assert "'b'" in str(exc_info.value)
        with pytest.raises(TypeError) as exc_info:
            _ = t[f1(c=1.0, a=t.c2)].show(0)
        assert "'b'" in str(exc_info.value)

        # bad default value
        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf(return_type=IntType(), param_types=[IntType(), FloatType(), FloatType()])
            def f1(a: int, b: float, c: str = '') -> float:
                return a + b + c
        assert 'default value' in str(exc_info.value).lower()
        # missing param type
        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf(return_type=IntType(), param_types=[IntType(), FloatType()])
            def f1(a: int, b: float, c: str = '') -> float:
                return a + b + c
        assert 'missing type for parameter c' in str(exc_info.value).lower()
        # bad parameter name
        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf(return_type=IntType(), param_types=[IntType()])
            def f1(group_by: int) -> int:
                return group_by
        assert 'reserved' in str(exc_info.value)
        # bad parameter name
        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf(return_type=IntType(), param_types=[IntType()])
            def f1(order_by: int) -> int:
                return order_by
        assert 'reserved' in str(exc_info.value)

    def test_expr_udf(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        @pxt.expr_udf
        def times2(x: int) -> int:
            return x + x
        res1 = t.select(out=times2(t.c2)).order_by(t.c2).collect()
        res2 = t.select(t.c2 * 2).order_by(t.c2).collect()
        assert_resultset_eq(res1, res2)

        with pytest.raises(TypeError) as exc_info:
            _ = t.select(times2(y=t.c2)).collect()
        assert 'missing a required argument' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # parameter types cannot be inferred
            @pxt.expr_udf
            def add1(x, y) -> int:
                return x + y
        assert 'cannot infer pixeltable type' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # return type cannot be inferred
            @pxt.expr_udf
            def add1(x: int, y: int):
                return x + y
        assert 'cannot infer pixeltable return type' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # missing param types
            @pxt.expr_udf(param_types=[IntType()])
            def add1(x, y) -> int:
                return x + y
        assert 'missing type for parameter y' in str(exc_info.value).lower()

        with pytest.raises(TypeError) as exc_info:
            # signature has correct parameter kind
            @pxt.expr_udf
            def add1(*, x: int) -> int:
                return x + 1
            _ = t.select(add1(t.c2)).collect()
        assert 'takes 0 positional arguments' in str(exc_info.value).lower()

        @pxt.expr_udf
        def add2(x: int, y: int = 1) -> int:
            return x + y
        res1 = t.select(out=add2(t.c2)).order_by(t.c2).collect()

    # Test that various invalid udf definitions generate
    # correct error messages.
    def test_invalid_udfs(self):
        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf
            def udf1(name: Batch[str]) -> str:
                return ''
        assert 'batched parameters in udf, but no `batch_size` given' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf(batch_size=32)
            def udf2(name: Batch[str]) -> str:
                return ''
        assert 'batch_size is specified; Python return type must be a `Batch`' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf
            def udf3(name: str) -> Optional[np.ndarray]:
                return None
        assert 'cannot infer pixeltable return type' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf
            def udf4(array: np.ndarray) -> str:
                return ''
        assert 'cannot infer pixeltable type for parameter array' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            @pxt.udf
            def udf5(name: str, untyped) -> str:
                return ''
        assert 'cannot infer pixeltable type for parameter untyped' in str(exc_info.value).lower()




from typing import Any

import pixeltable.func as func


def help(obj: Any) -> None:
    """Returns help text for the given object."""
    if isinstance(obj, func.Function):
        print(obj.help_str())
    else:
        print(__builtins__.help(obj))



from typing import Optional, List

import sqlalchemy as sql

import pixeltable.exprs as exprs

class ExecContext:
    """Class for execution runtime constants"""
    def __init__(
            self, row_builder: exprs.RowBuilder, *, show_pbar: bool = False, batch_size: int = 0,
            pk_clause: Optional[List[sql.ClauseElement]] = None, num_computed_exprs: int = 0,
            ignore_errors: bool = False
    ):
        self.show_pbar = show_pbar
        self.batch_size = batch_size
        self.profile = exprs.ExecProfile(row_builder)
        # num_rows is used to compute the total number of computed cells used for the progress bar
        self.num_rows: Optional[int] = None
        self.conn: Optional[sql.engine.Connection] = None  # if present, use this to execute SQL queries
        self.pk_clause = pk_clause
        self.num_computed_exprs = num_computed_exprs
        self.ignore_errors = ignore_errors



from typing import Generator, Optional

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode
import pixeltable.catalog as catalog
import pixeltable.exprs as exprs
import pixeltable.exceptions as excs


class ComponentIterationNode(ExecNode):
    """Expands each row from a base table into one row per component returned by an iterator

    Returns row batches of OUTPUT_BATCH_SIZE size.
    """
    OUTPUT_BATCH_SIZE = 1024

    def __init__(self, view: catalog.TableVersion, input: ExecNode):
        assert view.is_component_view()
        super().__init__(input.row_builder, [], [], input)
        self.view = view
        iterator_args = [view.iterator_args.copy()]
        self.row_builder.substitute_exprs(iterator_args)
        self.iterator_args = iterator_args[0]
        assert isinstance(self.iterator_args, exprs.InlineDict)
        self.iterator_args_ctx = self.row_builder.create_eval_ctx([self.iterator_args])
        self.iterator_output_schema, self.unstored_column_names = \
            self.view.iterator_cls.output_schema(**self.iterator_args.to_dict())
        self.iterator_output_fields = list(self.iterator_output_schema.keys())
        self.iterator_output_cols = \
            {field_name: self.view.cols_by_name[field_name] for field_name in self.iterator_output_fields}
        # referenced iterator output fields
        self.refd_output_slot_idxs = {
            e.col.name: e.slot_idx for e in self.row_builder.unique_exprs
            if isinstance(e, exprs.ColumnRef) and e.col.name in self.iterator_output_fields
        }
        self._output: Optional[Generator[DataRowBatch, None, None]] = None

    def _output_batches(self) -> Generator[DataRowBatch, None, None]:
        output_batch = DataRowBatch(self.view, self.row_builder)
        for input_batch in self.input:
            for input_row in input_batch:
                self.row_builder.eval(input_row, self.iterator_args_ctx)
                iterator_args = input_row[self.iterator_args.slot_idx]
                iterator = self.view.iterator_cls(**iterator_args)
                for pos, component_dict in enumerate(iterator):
                    output_row = output_batch.add_row()
                    input_row.copy(output_row)
                    # we're expanding the input and need to add the iterator position to the pk
                    pk = output_row.pk[:-1] + (pos,) + output_row.pk[-1:]
                    output_row.set_pk(pk)

                    # verify and copy component_dict fields to their respective slots in output_row
                    for field_name, field_val in component_dict.items():
                        if field_name not in self.iterator_output_fields:
                            raise excs.Error(
                                f'Invalid field name {field_name} in output of {self.view.iterator_cls.__name__}')
                        if field_name not in self.refd_output_slot_idxs:
                            # we can ignore this
                            continue
                        output_col = self.iterator_output_cols[field_name]
                        output_col.col_type.validate_literal(field_val)
                        output_row[self.refd_output_slot_idxs[field_name]] = field_val
                    if len(component_dict) != len(self.iterator_output_fields):
                        missing_fields = set(self.refd_output_slot_idxs.keys()) - set(component_dict.keys())
                        raise excs.Error(
                            f'Invalid output of {self.view.iterator_cls.__name__}: '
                            f'missing fields {", ".join(missing_fields)}')

                    if len(output_batch) == self.OUTPUT_BATCH_SIZE:
                        yield output_batch
                        output_batch = DataRowBatch(self.view, self.row_builder)

        if len(output_batch) > 0:
            yield output_batch

    def __next__(self) -> DataRowBatch:
        if self._output is None:
            self._output = self._output_batches()
        return next(self._output)



from .arithmetic_expr import ArithmeticExpr
from .array_slice import ArraySlice
from .column_property_ref import ColumnPropertyRef
from .column_ref import ColumnRef
from .comparison import Comparison
from .compound_predicate import CompoundPredicate
from .data_row import DataRow
from .expr import Expr
from .function_call import FunctionCall
from .image_member_access import ImageMemberAccess
from .image_similarity_predicate import ImageSimilarityPredicate
from .inline_array import InlineArray
from .inline_dict import InlineDict
from .is_null import IsNull
from .json_mapper import JsonMapper
from .json_path import RELATIVE_PATH_ROOT, JsonPath
from .literal import Literal
from .object_ref import ObjectRef
from  .variable import Variable
from .predicate import Predicate
from .row_builder import RowBuilder, ColumnSlotIdx, ExecProfile
from .rowid_ref import RowidRef
from .expr_set import ExprSet
from .type_cast import TypeCast



import numpy as np
import PIL.Image

import pixeltable.func as func
from pixeltable.env import Env


def embed_image(img: PIL.Image.Image) -> np.ndarray:
    from pixeltable.functions.nos.image_embedding import openai_clip
    model_info = openai_clip.model_spec
    result = Env.get().nos_client.Run(task=model_info.task, model_name=model_info.name, images=[img.resize((224, 224))])
    return result['embedding'].squeeze(0)

def embed_text(text: str) -> np.ndarray:
    from pixeltable.functions.nos.text_embedding import openai_clip
    model_info = openai_clip.model_spec
    result = Env.get().nos_client.Run(task=model_info.task, model_name=model_info.name, texts=[text])
    return result['embedding'].squeeze(0)



from __future__ import annotations
from typing import Optional, List, Tuple, Callable

from .expr import Expr
from .globals import LogicalOperator
import pixeltable
import pixeltable.type_system as ts


class Predicate(Expr):
    def __init__(self) -> None:
        super().__init__(ts.BoolType())

    def split_conjuncts(
            self, condition: Callable[[Predicate], bool]) -> Tuple[List[Predicate], Optional[Predicate]]:
        """
        Returns clauses of a conjunction that meet condition in the first element.
        The second element contains remaining clauses, rolled into a conjunction.
        """
        if condition(self):
            return [self], None
        else:
            return [], self

    def __and__(self, other: object) -> 'pixeltable.exprs.CompoundPredicate':
        if not isinstance(other, Expr):
            raise TypeError(f'Other needs to be an expression: {type(other)}')
        if not other.col_type.is_bool_type():
            raise TypeError(f'Other needs to be an expression that returns a boolean: {other.col_type}')
        from .compound_predicate import CompoundPredicate
        return CompoundPredicate(LogicalOperator.AND, [self, other])

    def __or__(self, other: object) -> 'pixeltable.exprs.CompoundPredicate':
        if not isinstance(other, Expr):
            raise TypeError(f'Other needs to be an expression: {type(other)}')
        if not other.col_type.is_bool_type():
            raise TypeError(f'Other needs to be an expression that returns a boolean: {other.col_type}')
        from .compound_predicate import CompoundPredicate
        return CompoundPredicate(LogicalOperator.OR, [self, other])

    def __invert__(self) -> 'pixeltable.exprs.CompoundPredicate':
        from .compound_predicate import CompoundPredicate
        return CompoundPredicate(LogicalOperator.NOT, [self])




import json
import PIL.Image
import io
import pyarrow.parquet as pq
import pyarrow as pa
import numpy as np
from pathlib import Path
from collections import deque

from typing import List, Tuple, Any, Dict
from pixeltable.utils.transactional_directory import transactional_directory

def _write_batch(value_batch : Dict[str, deque], schema : pa.Schema, output_path : Path) -> None:
    pydict = {}
    for field in schema:
        if isinstance(field.type, pa.FixedShapeTensorType):
            stacked_arr = np.stack(value_batch[field.name])
            pydict[field.name] = pa.FixedShapeTensorArray.from_numpy_ndarray(stacked_arr)
        else:
            pydict[field.name] = value_batch[field.name]

    tab = pa.Table.from_pydict(pydict, schema=schema)
    pq.write_table(tab, output_path)

def save_parquet(df: 'pixeltable.DataFrame', dest_path: Path, partition_size_bytes: int = 100_000_000) -> None:
    """
        Internal method to stream dataframe data to parquet format.
        Does not materialize the dataset to memory.

        It preserves pixeltable type metadata in a json file, which would otherwise 
        not be available in the parquet format.

        Images are stored inline in a compressed format in their parquet file.
    
        Args:
            df : dataframe to save.
            dest_path : path to directory to save the parquet files to.
            partition_size_bytes : maximum target size for each chunk. Default 100_000_000 bytes.
    """
    column_names = df.get_column_names()
    column_types = df.get_column_types()

    type_dict = {k: v.as_dict() for k, v in zip(column_names, column_types)}

    arrow_schema = pa.schema([pa.field(name, column_types[i].to_arrow_type()) for i, name in enumerate(column_names)])
    # store the changes atomically
    with transactional_directory(dest_path) as temp_path:
        # dump metadata json file so we can inspect what was the source of the parquet file later on.
        json.dump(df._as_dict(), (temp_path / '.pixeltable.json').open('w')) # pylint: disable=protected-access
        json.dump(type_dict, (temp_path / '.pixeltable.column_types.json').open('w')) # keep type metadata

        batch_num = 0
        current_value_batch : Dict[str, deque] = {k:deque() for k in column_names}
        current_byte_estimate = 0

        for data_row in df._exec(): # pylint: disable=protected-access
            for (col_name, col_type, e) in zip(column_names, column_types, df._select_list_exprs): # pylint: disable=protected-access
                val = data_row[e.slot_idx]
                if val is None:
                    current_value_batch[col_name].append(val)
                    continue

                assert val is not None
                if col_type.is_image_type():
                    # images get inlined into the parquet file
                    if data_row.file_paths is not None and data_row.file_paths[e.slot_idx] is not None:
                        # if there is a file, read directly to preserve information
                        val = open(data_row.file_paths[e.slot_idx], 'rb').read()
                    elif isinstance(val, PIL.Image.Image): 
                        # if no file available, eg. bc it is computed, convert to png
                        buf = io.BytesIO()
                        val.save(buf, format='PNG')
                        val = buf.getvalue()
                    else:
                        assert False, f'unknown image type {type(val)}'
                    length = len(val)
                elif col_type.is_string_type():
                    length = len(val)
                elif col_type.is_video_type():
                    if data_row.file_paths is not None and data_row.file_paths[e.slot_idx] is not None:
                        val = data_row.file_paths[e.slot_idx]
                    else:
                        assert False, f'unknown video type {type(val)}'
                    length = len(val)
                elif col_type.is_json_type():
                    val = json.dumps(val)
                    length = len(val)
                elif col_type.is_array_type():
                    length = val.nbytes
                elif col_type.is_int_type():
                    length = 8
                elif col_type.is_float_type():
                    length = 8
                elif col_type.is_bool_type():
                    length = 1
                elif col_type.is_timestamp_type():
                    length = 8
                else:
                    assert False, f'unknown type {col_type} for {col_name}'

                current_value_batch[col_name].append(val)
                current_byte_estimate += length
            if current_byte_estimate > partition_size_bytes:
                assert batch_num < 100_000, 'wrote too many parquet files, unclear ordering'
                _write_batch(current_value_batch, arrow_schema,  temp_path / f'part-{batch_num:05d}.parquet')
                batch_num += 1
                current_value_batch = {k:deque() for k in column_names}
                current_byte_estimate = 0

        _write_batch(current_value_batch, arrow_schema, temp_path / f'part-{batch_num:05d}.parquet')

def get_part_metadata(path : Path) -> List[Tuple[str, int]]:
    """
        Args:
            path: path to directory containing parquet files.
        Returns:
            A list of (file_name, num_rows) tuples for the parquet files in file name order.
    """
    parts = sorted([f for f in path.iterdir() if f.suffix == '.parquet'])
    rows_per_file = {}

    for part in parts:
        parquet_file = pq.ParquetFile(str(part))
        rows_per_file[part] = parquet_file.metadata.num_rows

    return [(file, num_rows) for file, num_rows in rows_per_file.items()]


from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple, Union

import PIL
import sqlalchemy as sql

from .expr import Expr
from .column_ref import ColumnRef
from .function_call import FunctionCall
from .image_similarity_predicate import ImageSimilarityPredicate
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog
import pixeltable.func as func
import pixeltable.exceptions as excs
import pixeltable.type_system as ts


# TODO: this doesn't dig up all attrs for actual jpeg images
def _create_pil_attr_info() -> Dict[str, ts.ColumnType]:
    # create random Image to inspect for attrs
    img = PIL.Image.new('RGB', (100, 100))
    # we're only interested in public attrs (including properties)
    result: Dict[str, ts.ColumnType] = {}
    for name in [name for name in dir(img) if not callable(getattr(img, name)) and not name.startswith('_')]:
        if getattr(img, name) is None:
            continue
        if isinstance(getattr(img, name), str):
            result[name] = ts.StringType()
        if isinstance(getattr(img, name), int):
            result[name] = ts.IntType()
        if getattr(img, name) is dict:
            result[name] = ts.JsonType()
    return result


class ImageMemberAccess(Expr):
    """
    Access of either an attribute or function member of PIL.Image.Image.
    Ex.: tbl.img_col_ref.rotate(90), tbl.img_col_ref.width
    TODO: remove this class and use FunctionCall instead (attributes to be replaced by functions)
    """
    attr_info = _create_pil_attr_info()

    def __init__(self, member_name: str, caller: Expr):
        if member_name == 'nearest':
            super().__init__(ts.InvalidType())  # requires FunctionCall to return value
        elif member_name in self.attr_info:
            super().__init__(self.attr_info[member_name])
        else:
            candidates = func.FunctionRegistry.get().get_type_methods(member_name, ts.ColumnType.Type.IMAGE)
            if len(candidates) == 0:
                raise excs.Error(f'Unknown Image member: {member_name}')
            if len(candidates) > 1:
                raise excs.Error(f'Ambiguous Image method: {member_name}')
            self.img_method = candidates[0]
            super().__init__(ts.InvalidType())  # requires FunctionCall to return value
        self.member_name = member_name
        self.components = [caller]
        self.id = self._create_id()

    def default_column_name(self) -> Optional[str]:
        return self.member_name.replace('.', '_')

    @property
    def _caller(self) -> Expr:
        return self.components[0]

    def __str__(self) -> str:
        return f'{self._caller}.{self.member_name}'

    def _as_dict(self) -> Dict:
        return {'member_name': self.member_name, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'member_name' in d
        assert len(components) == 1
        return cls(d['member_name'], components[0])

    def __call__(self, *args, **kwargs) -> Union[FunctionCall, ImageSimilarityPredicate]:
        caller = self._caller
        call_signature = f'({",".join([type(arg).__name__ for arg in args])})'
        if self.member_name == 'nearest':
            # - caller must be ColumnRef
            # - signature is (Union[PIL.Image.Image, str])
            if not isinstance(caller, ColumnRef):
                raise excs.Error(f'nearest(): caller must be an image column')
            if len(args) != 1 or (not isinstance(args[0], PIL.Image.Image) and not isinstance(args[0], str)):
                raise excs.Error(f'nearest(): requires a PIL.Image.Image or str, got {call_signature} instead')
            return ImageSimilarityPredicate(
                caller,
                img=args[0] if isinstance(args[0], PIL.Image.Image) else None,
                text=args[0] if isinstance(args[0], str) else None)

        result = self.img_method(*[caller, *args], **kwargs)
        result.is_method_call = True
        return result

    def _equals(self, other: ImageMemberAccess) -> bool:
        return self.member_name == other.member_name

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('member_name', self.member_name)]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        caller_val = data_row[self._caller.slot_idx]
        try:
            data_row[self.slot_idx] = getattr(caller_val, self.member_name)
        except AttributeError:
            data_row[self.slot_idx] = None




from typing import Dict, Any, Tuple, Optional

import PIL.Image

from pixeltable.type_system import FloatType, ImageType, IntType, ArrayType, ColumnType, StringType, JsonType, BoolType
import pixeltable.func as func


def _caller_return_type(bound_args: Optional[Dict[str, Any]]) -> ColumnType:
    if bound_args is None:
        return ImageType()
    return bound_args['self'].col_type

@func.udf(
    py_fn=PIL.Image.alpha_composite, return_type=ImageType(), param_types=[ImageType(), ImageType()])
def alpha_composite(im1: PIL.Image.Image, im2: PIL.Image.Image) -> PIL.Image.Image:
    pass
@func.udf(
    py_fn=PIL.Image.blend, return_type=ImageType(), param_types=[ImageType(), ImageType(), FloatType()])
def blend(im1: PIL.Image.Image, im2: PIL.Image.Image, alpha: float) -> PIL.Image.Image:
    pass
@func.udf(
    py_fn=PIL.Image.composite, return_type=ImageType(), param_types=[ImageType(), ImageType(), ImageType()])
def composite(image1: PIL.Image.Image, image2: PIL.Image.Image, mask: PIL.Image.Image) -> PIL.Image.Image:
    pass


# PIL.Image.Image methods

# Image.convert()
def _convert_return_type(bound_args: Dict[str, Any]) -> ColumnType:
    if bound_args is None:
        return ImageType()
    assert 'self' in bound_args
    assert 'mode' in bound_args
    img_type = bound_args['self'].col_type
    return ImageType(size=img_type.size, mode=bound_args['mode'])
@func.udf(return_type=_convert_return_type, param_types=[ImageType(), StringType()])
def convert(self: PIL.Image.Image, mode: str) -> PIL.Image.Image:
    return self.convert(mode)

# Image.crop()
def _crop_return_type(bound_args: Dict[str, Any]) -> ColumnType:
    if bound_args is None:
        return ImageType()
    img_type = bound_args['self'].col_type
    box = bound_args['box']
    if isinstance(box, list) and all(isinstance(x, int) for x in box):
        return ImageType(size=(box[2] - box[0], box[3] - box[1]), mode=img_type.mode)
    return ImageType()  # we can't compute the size statically
@func.udf(
    py_fn=PIL.Image.Image.crop, return_type=_crop_return_type,
    param_types=[ImageType(), ArrayType((4,), dtype=IntType())])
def crop(self: PIL.Image.Image, box: Tuple[int, int, int, int]) -> PIL.Image.Image:
    pass

# Image.getchannel()
def _getchannel_return_type(bound_args: Dict[str, Any]) -> ColumnType:
    if bound_args is None:
        return ImageType()
    img_type = bound_args['self'].col_type
    return ImageType(size=img_type.size, mode='L')
@func.udf(
    py_fn=PIL.Image.Image.getchannel, return_type=_getchannel_return_type, param_types=[ImageType(), IntType()])
def getchannel(self: PIL.Image.Image, channel: int) -> PIL.Image.Image:
    pass

# Image.resize()
def resize_return_type(bound_args: Dict[str, Any]) -> ColumnType:
    if bound_args is None:
        return ImageType()
    assert 'size' in bound_args
    return ImageType(size=bound_args['size'])
@func.udf(return_type=resize_return_type, param_types=[ImageType(), ArrayType((2, ), dtype=IntType())])
def resize(self: PIL.Image.Image, size: Tuple[int, int]) -> PIL.Image.Image:
    return self.resize(size)

# Image.rotate()
@func.udf(return_type=ImageType(), param_types=[ImageType(), IntType()])
def rotate(self: PIL.Image.Image, angle: int) -> PIL.Image.Image:
    return self.rotate(angle)

# Image.transform()
@func.udf(return_type= _caller_return_type, param_types=[ImageType(), ArrayType((2,), dtype=IntType()), IntType()])
def transform(self: PIL.Image.Image, size: Tuple[int, int], method: int) -> PIL.Image.Image:
    return self.transform(size, method)

@func.udf(
    py_fn=PIL.Image.Image.effect_spread, return_type=_caller_return_type, param_types=[ImageType(), FloatType()])
def effect_spread(self: PIL.Image.Image, distance: float) -> PIL.Image.Image:
    pass

@func.udf(
    py_fn=PIL.Image.Image.entropy, return_type=FloatType(), param_types=[ImageType(), ImageType(), JsonType()])
def entropy(self: PIL.Image.Image, mask: PIL.Image.Image, histogram: Dict) -> float:
    pass

@func.udf(py_fn=PIL.Image.Image.getbands, return_type=JsonType(), param_types=[ImageType()])
def getbands(self: PIL.Image.Image) -> Tuple[str]:
    pass

@func.udf(py_fn=PIL.Image.Image.getbbox, return_type=JsonType(), param_types=[ImageType()])
def getbbox(self: PIL.Image.Image) -> Tuple[int, int, int, int]:
    pass

@func.udf(
    py_fn=PIL.Image.Image.getcolors, return_type=JsonType(), param_types=[ImageType(), IntType()])
def getcolors(self: PIL.Image.Image, maxcolors: int) -> Tuple[Tuple[int, int, int], int]:
    pass

@func.udf(py_fn=PIL.Image.Image.getextrema, return_type=JsonType(), param_types=[ImageType()])
def getextrema(self: PIL.Image.Image) -> Tuple[int, int]:
    pass

@func.udf(
    py_fn=PIL.Image.Image.getpalette, return_type=JsonType(), param_types=[ImageType(), StringType()])
def getpalette(self: PIL.Image.Image, mode: str) -> Tuple[int]:
    pass

@func.udf(
    py_fn=PIL.Image.Image.getpixel, return_type=JsonType(), param_types=[ImageType(), ArrayType((2,), dtype=IntType())])
def getpixel(self: PIL.Image.Image, xy: Tuple[int, int]) -> Tuple[int]:
    pass

@func.udf(
    py_fn=PIL.Image.Image.getprojection, return_type=JsonType(), param_types=[ImageType()])
def getprojection(self: PIL.Image.Image) -> Tuple[int]:
    pass

@func.udf(
    py_fn=PIL.Image.Image.histogram, return_type=JsonType(), param_types=[ImageType(), ImageType(), JsonType()])
def histogram(self: PIL.Image.Image, mask: PIL.Image.Image, histogram: Dict) -> Tuple[int]:
    pass

@func.udf(
    py_fn=PIL.Image.Image.quantize, return_type=ImageType(),
    param_types=[ImageType(), IntType(), IntType(nullable=True), IntType(), IntType(nullable=True), IntType()])
def quantize(
        self: PIL.Image.Image, colors: int, method: int, kmeans: int, palette: int, dither: int) -> PIL.Image.Image:
    pass

@func.udf(
    py_fn=PIL.Image.Image.reduce, return_type=ImageType(), param_types=[ImageType(), IntType(), JsonType()])
def reduce(self: PIL.Image.Image, factor: int, filter: Tuple[int]) -> PIL.Image.Image:
    pass

@func.udf(
    py_fn=PIL.Image.Image.transpose, return_type=_caller_return_type, param_types=[ImageType(), IntType()])
def transpose(self: PIL.Image.Image, method: int) -> PIL.Image.Image:
    pass



from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple

import sqlalchemy as sql
import PIL
import numpy as np

from .expr import Expr
from .predicate import Predicate
from .column_ref import ColumnRef
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog
import pixeltable.utils.clip as clip

class ImageSimilarityPredicate(Predicate):
    def __init__(self, img_col_ref: ColumnRef, img: Optional[PIL.Image.Image] = None, text: Optional[str] = None):
        assert (img is None) != (text is None)
        super().__init__()
        self.img_col_ref = img_col_ref
        self.components = [img_col_ref]
        self.img = img
        self.text = text
        self.id = self._create_id()

    def embedding(self) -> np.ndarray:
        if self.text is not None:
            return clip.embed_text(self.text)
        else:
            return clip.embed_image(self.img)

    def __str__(self) -> str:
        return f'{str(self.img_col_ref)}.nearest({"<img>" if self.img is not None else self.text})'

    def _equals(self, other: ImageSimilarityPredicate) -> bool:
        return False

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('img', id(self.img)), ('text', self.text)]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        assert False

    def _as_dict(self) -> Dict:
        assert False, 'not implemented'
        # TODO: convert self.img into a serializable string
        return {'img': self.img, 'text': self.text, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'img' in d
        assert 'text' in d
        assert len(components) == 1
        return cls(components[0], d['img'], d['text'])




from __future__ import annotations
from typing import Iterable, Optional

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode
import pixeltable.exprs as exprs
import pixeltable.exceptions as excs


class MediaValidationNode(ExecNode):
    """Validation of selected media slots
    Records exceptions in the rows of the input batch
    """
    def __init__(
            self, row_builder: exprs.RowBuilder, media_slots: Iterable[exprs.ColumnSlotIdx],
            input: Optional[ExecNode]):
        super().__init__(row_builder, [], [], input)
        self.row_builder = row_builder
        self.input = input
        for col in [c.col for c in media_slots]:
            assert col.col_type.is_media_type()
        self.media_slots = media_slots

    def __next__(self) -> DataRowBatch:
        assert self.input is not None
        row_batch = next(self.input)
        for row in row_batch:
            for slot_idx, col in [(c.slot_idx, c.col) for c in self.media_slots]:
                if row.has_exc(slot_idx):
                    continue
                assert row.has_val[slot_idx]
                path = row.file_paths[slot_idx]
                if path is None:
                    continue

                try:
                    col.col_type.validate_media(path)
                except excs.Error as exc:
                    self.row_builder.set_exc(row, slot_idx, exc)
                    if not self.ctx.ignore_errors:
                        raise exc

        return row_batch



from typing import Optional, List, Dict, get_type_hints, Type, Any, TypeVar, Tuple, Union
import platform
import uuid
import dataclasses

import sqlalchemy as sql
from sqlalchemy import Integer, String, Boolean, BigInteger, LargeBinary
from sqlalchemy.dialects.postgresql import UUID, JSONB
from sqlalchemy import ForeignKey, UniqueConstraint, ForeignKeyConstraint
from sqlalchemy.orm import declarative_base

Base = declarative_base()

T = TypeVar('T')

def md_from_dict(data_class_type: Type[T], data: Any) -> T:
    """Re-instantiate a dataclass instance that contains nested dataclasses from a dict."""
    if dataclasses.is_dataclass(data_class_type):
        fieldtypes = {f: t for f, t in get_type_hints(data_class_type).items()}
        return data_class_type(**{f: md_from_dict(fieldtypes[f], data[f]) for f in data})
    elif hasattr(data_class_type, '__origin__'):
        if data_class_type.__origin__ is Union and type(None) in data_class_type.__args__:
            # Handling Optional types
            non_none_args = [arg for arg in data_class_type.__args__ if arg is not type(None)]
            if len(non_none_args) == 1:
                return md_from_dict(non_none_args[0], data) if data is not None else None
        elif data_class_type.__origin__ is list:
            return [md_from_dict(data_class_type.__args__[0], elem) for elem in data]
        elif data_class_type.__origin__ is dict:
            key_type = data_class_type.__args__[0]
            val_type = data_class_type.__args__[1]
            return {key_type(key): md_from_dict(val_type, val) for key, val in data.items()}
        elif data_class_type.__origin__ is tuple:
            return tuple(md_from_dict(arg_type, elem) for arg_type, elem in zip(data_class_type.__args__, data))
    else:
        return data


# structure of the stored metadata:
# - each schema entity that grows somehow proportionally to the data (# of output_rows, total insert operations,
#   number of schema changes) gets its own table
# - each table has an 'md' column that basically contains the payload
# - exceptions to that are foreign keys without which lookups would be too slow (ex.: TableSchemaVersions.tbl_id)
# - the md column contains a dataclass serialized to json; this has the advantage of making changes to the metadata
#   schema easier (the goal is not to have to rely on some schema migration framework; if that breaks for some user,
#   it would be very difficult to patch up)

@dataclasses.dataclass
class SystemInfoMd:
    schema_version: int


class SystemInfo(Base):
    """A single-row table that contains system-wide metadata."""
    __tablename__ = 'systeminfo'
    dummy = sql.Column(Integer, primary_key=True, default=0, nullable=False)
    md = sql.Column(JSONB, nullable=False)  # SystemInfoMd


@dataclasses.dataclass
class DirMd:
    name: str


class Dir(Base):
    __tablename__ = 'dirs'

    id = sql.Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4, nullable=False)
    parent_id = sql.Column(UUID(as_uuid=True), ForeignKey('dirs.id'), nullable=True)
    md = sql.Column(JSONB, nullable=False)


@dataclasses.dataclass
class ColumnHistory:
    """
    Records when a column was added/dropped, which is needed to GC unreachable storage columns
    (a column that was added after table snapshot n and dropped before table snapshot n+1 can be removed
    from the stored table).
    One record per column (across all schema versions).
     """
    col_id: int
    schema_version_add: int
    schema_version_drop: Optional[int]


@dataclasses.dataclass
class ViewMd:
    is_snapshot: bool

    # (table id, version); for mutable views, all versions are None
    base_versions: List[Tuple[str, Optional[int]]]

    # filter predicate applied to the base table; view-only
    predicate: Optional[Dict[str, Any]]

    # ComponentIterator subclass; only for component views
    iterator_class_fqn: Optional[str]

    # args to pass to the iterator class constructor; only for component views
    iterator_args: Optional[Dict[str, Any]]


@dataclasses.dataclass
class TableMd:
    name: str

    # monotonically increasing w/in Table for both data and schema changes, starting at 0
    current_version: int
    # each version has a corresponding schema version (current_version >= current_schema_version)
    current_schema_version: int

    # used to assign Column.id
    next_col_id: int

    # - used to assign the rowid column in the storage table
    # - every row is assigned a unique and immutable rowid on insertion
    next_row_id: int

    column_history: Dict[int, ColumnHistory]  # col_id -> ColumnHistory

    view_md: Optional[ViewMd]


class Table(Base):
    """
    Table represents both tables and views.

    Views are in essence a subclass of tables, because they also store materialized columns. The differences are:
    - views have a base, which is either a (live) table or a snapshot
    - views can have a filter predicate
    """
    __tablename__ = 'tables'

    MAX_VERSION = 9223372036854775807  # 2^63 - 1

    id = sql.Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4, nullable=False)
    dir_id = sql.Column(UUID(as_uuid=True), ForeignKey('dirs.id'), nullable=False)
    md = sql.Column(JSONB, nullable=False)  # TableMd


@dataclasses.dataclass
class TableVersionMd:
    created_at: float  # time.time()
    version: int
    schema_version: int


class TableVersion(Base):
    __tablename__ = 'tableversions'
    tbl_id = sql.Column(UUID(as_uuid=True), ForeignKey('tables.id'), primary_key=True, nullable=False)
    version = sql.Column(BigInteger, primary_key=True, nullable=False)
    md = sql.Column(JSONB, nullable=False)  # TableVersionMd


@dataclasses.dataclass
class SchemaColumn:
    """
    Records the logical (user-visible) schema of a table.
    Contains the full set of columns for each new schema version: one record per (column x schema version).
    """
    pos: int
    name: str
    col_type: dict
    is_pk: bool
    value_expr: Optional[dict]
    stored: Optional[bool]
    # if True, creates vector index for this column
    is_indexed: bool


@dataclasses.dataclass
class TableSchemaVersionMd:
    schema_version: int
    preceding_schema_version: Optional[int]
    columns: Dict[int, SchemaColumn]  # col_id -> SchemaColumn
    num_retained_versions: int
    comment: str


# versioning: each table schema change results in a new record
class TableSchemaVersion(Base):
    __tablename__ = 'tableschemaversions'

    tbl_id = sql.Column(UUID(as_uuid=True), ForeignKey('tables.id'), primary_key=True, nullable=False)
    schema_version = sql.Column(BigInteger, primary_key=True, nullable=False)
    md = sql.Column(JSONB, nullable=False)  # TableSchemaVersionMd


@dataclasses.dataclass
class FunctionMd:
    name: str
    py_version: str  # platform.python_version
    class_name: str  # name of the Function subclass
    md: dict  # part of the output of Function.to_store()


class Function(Base):
    """
    User-defined functions that are not module functions (ie, aren't available at runtime as a symbol in a known
    module).
    Functions without a name are anonymous functions used in the definition of a computed column.
    Functions that have names are also assigned to a database and directory.
    We store the Python version under which a Function was created (and the callable pickled) in order to warn
    against version mismatches.
    """
    __tablename__ = 'functions'

    id = sql.Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4, nullable=False)
    dir_id = sql.Column(UUID(as_uuid=True), ForeignKey('dirs.id'), nullable=True)
    md = sql.Column(JSONB, nullable=False)  # FunctionMd
    binary_obj = sql.Column(LargeBinary, nullable=True)



from __future__ import annotations

import inspect
import sys
from typing import Optional, Dict, Callable, List, Tuple
from uuid import UUID
import cloudpickle

import pixeltable.type_system as ts
import pixeltable.exceptions as excs
from .function import Function
from .function_registry import FunctionRegistry
from .globals import get_caller_module_path
from .signature import Signature


class CallableFunction(Function):
    """Pixeltable Function backed by a Python Callable.

    CallableFunctions come in two flavors:
    - references to lambdas and functions defined in notebooks, which are pickled and serialized to the store
    - functions that are defined in modules are serialized via the default mechanism
    """

    def __init__(
            self, signature: Signature, py_fn: Callable, self_path: Optional[str] = None,
            self_name: Optional[str] = None):
        assert py_fn is not None
        self.py_fn = py_fn
        self.self_name = self_name
        py_signature = inspect.signature(self.py_fn)
        super().__init__(signature, py_signature, self_path=self_path)

    @property
    def display_name(self) -> str:
        return self.self_name

    @property
    def name(self) -> str:
        return self.self_name

    def help_str(self) -> str:
        res = super().help_str()
        res += '\n\n' + inspect.getdoc(self.py_fn)
        return res

    def _as_dict(self) -> Dict:
        if self.self_path is None:
            # this is not a module function
            from .function_registry import FunctionRegistry
            id = FunctionRegistry.get().create_stored_function(self)
            return {'id': id.hex}
        return super()._as_dict()

    @classmethod
    def _from_dict(cls, d: Dict) -> Function:
        if 'id' in d:
            from .function_registry import FunctionRegistry
            return FunctionRegistry.get().get_stored_function(UUID(hex=d['id']))
        return super()._from_dict(d)

    def to_store(self) -> Tuple[Dict, bytes]:
        return (self.signature.as_dict(), cloudpickle.dumps(self.py_fn))

    @classmethod
    def from_store(cls, name: Optional[str], md: Dict, binary_obj: bytes) -> Function:
        py_fn = cloudpickle.loads(binary_obj)
        assert isinstance(py_fn, Callable)
        return CallableFunction(Signature.from_dict(md), py_fn, self_name=name)



from typing import Optional, Dict
import dataclasses

import pixeltable.type_system as ts


@dataclasses.dataclass
class DocumentHandle:
    format: ts.DocumentType.DocumentFormat
    bs_doc: Optional['bs4.BeautifulSoup'] = None
    md_ast: Optional[Dict] = None


def get_document_handle(s: str) -> Optional[DocumentHandle]:
    bs_doc = get_html_handle(s)
    if bs_doc is not None:
        return DocumentHandle(format=ts.DocumentType.DocumentFormat.HTML, bs_doc=bs_doc)
    md_ast = get_markdown_handle(s)
    if md_ast is not None:
        return DocumentHandle(format=ts.DocumentType.DocumentFormat.MD, md_ast=md_ast)
    return None

def get_html_handle(s: str) -> Optional['bs4.BeautifulSoup']:
    import bs4
    try:
        doc = bs4.BeautifulSoup(s, 'html.parser')
    except Exception as e:
        return None
    if doc.find() is None:
        return None
    return doc

def get_markdown_handle(s: str) -> Optional[Dict]:
    import mistune
    try:
        md_ast = mistune.create_markdown(renderer=None)
        return md_ast(s)
    except Exception as e:
        return None



from __future__ import annotations

import abc
import datetime
import enum
import json
import typing
import urllib.parse
from copy import copy
from pathlib import Path
from typing import Any, Optional, Tuple, Dict, Callable, List, Union

import PIL.Image
import av
import numpy as np
import sqlalchemy as sql

from pixeltable import exceptions as excs


class ColumnType:
    @enum.unique
    class Type(enum.Enum):
        STRING = 0
        INT = 1
        FLOAT = 2
        BOOL = 3
        TIMESTAMP = 4
        JSON = 5
        ARRAY = 6
        IMAGE = 7
        VIDEO = 8
        AUDIO = 9
        DOCUMENT = 10

        # exprs that don't evaluate to a computable value in Pixeltable, such as an Image member function
        INVALID = 255

        @classmethod
        def supertype(
                cls, type1: 'Type', type2: 'Type',
                # we need to pass this in because we can't easily append it as a class member
                common_supertypes: Dict[Tuple['Type', 'Type'], 'Type']
        ) -> Optional['Type']:
            if type1 == type2:
                return type1
            t = common_supertypes.get((type1, type2))
            if t is not None:
                return t
            t = common_supertypes.get((type2, type1))
            if t is not None:
                return t
            return None


    @enum.unique
    class DType(enum.Enum):
        """
        Base type used in images and arrays
        """
        BOOL = 0,
        INT8 = 1,
        INT16 = 2,
        INT32 = 3,
        INT64 = 4,
        UINT8 = 5,
        UINT16 = 6,
        UINT32 = 7,
        UINT64 = 8,
        FLOAT16 = 9,
        FLOAT32 = 10,
        FLOAT64 = 11

    scalar_types = {Type.STRING, Type.INT, Type.FLOAT, Type.BOOL, Type.TIMESTAMP}
    numeric_types = {Type.INT, Type.FLOAT}
    common_supertypes: Dict[Tuple[Type, Type], Type] = {
        (Type.BOOL, Type.INT): Type.INT,
        (Type.BOOL, Type.FLOAT): Type.FLOAT,
        (Type.INT, Type.FLOAT): Type.FLOAT,
    }

    def __init__(self, t: Type, nullable: bool = False):
        self._type = t
        self.nullable = nullable

    @property
    def type_enum(self) -> Type:
        return self._type

    def serialize(self) -> str:
        return json.dumps(self.as_dict())

    @classmethod
    def serialize_list(cls, type_list: List[ColumnType]) -> str:
        return json.dumps([t.as_dict() for t in type_list])

    def as_dict(self) -> Dict:
        return {
            '_classname': self.__class__.__name__,
            **self._as_dict(),
        }

    def _as_dict(self) -> Dict:
        return {'nullable': self.nullable}

    @classmethod
    def deserialize(cls, type_str: str) -> ColumnType:
        type_dict = json.loads(type_str)
        return cls.from_dict(type_dict)

    @classmethod
    def deserialize_list(cls, type_list_str: str) -> List[ColumnType]:
        type_dict_list = json.loads(type_list_str)
        return [cls.from_dict(type_dict) for type_dict in type_dict_list]

    @classmethod
    def from_dict(cls, type_dict: Dict) -> ColumnType:
        assert '_classname' in type_dict
        type_class = globals()[type_dict['_classname']]
        return type_class._from_dict(type_dict)

    @classmethod
    def _from_dict(cls, d: Dict) -> ColumnType:
        """
        Default implementation: simply invoke c'tor
        """
        assert 'nullable' in d
        return cls(nullable=d['nullable'])

    @classmethod
    def make_type(cls, t: Type) -> ColumnType:
        assert t != cls.Type.INVALID and t != cls.Type.ARRAY
        if t == cls.Type.STRING:
            return StringType()
        if t == cls.Type.INT:
            return IntType()
        if t == cls.Type.FLOAT:
            return FloatType()
        if t == cls.Type.BOOL:
            return BoolType()
        if t == cls.Type.TIMESTAMP:
            return TimestampType()
        if t == cls.Type.JSON:
            return JsonType()
        if t == cls.Type.IMAGE:
            return ImageType()
        if t == cls.Type.VIDEO:
            return VideoType()
        if t == cls.Type.AUDIO:
            return AudioType()
        if t == cls.Type.DOCUMENT:
            return AudioType()

    def __str__(self) -> str:
        return self._type.name.lower()

    def __eq__(self, other: object) -> bool:
        return self.matches(other) and self.nullable == other.nullable

    def is_supertype_of(self, other: ColumnType) -> bool:
        if type(self) != type(other):
            return False
        if self.matches(other):
            return True
        return self._is_supertype_of(other)

    @abc.abstractmethod
    def _is_supertype_of(self, other: ColumnType) -> bool:
        return False

    def matches(self, other: object) -> bool:
        """Two types match if they're equal, aside from nullability"""
        if not isinstance(other, ColumnType):
            pass
        assert isinstance(other, ColumnType)
        if type(self) != type(other):
            return False
        for member_var in vars(self).keys():
            if member_var == 'nullable':
                continue
            if getattr(self, member_var) != getattr(other, member_var):
                return False
        return True

    @classmethod
    def supertype(cls, type1: ColumnType, type2: ColumnType) -> Optional[ColumnType]:
        if type1 == type2:
            return type1

        if type1.is_invalid_type():
            return type2
        if type2.is_invalid_type():
            return type1

        if type1.is_scalar_type() and type2.is_scalar_type():
            t = cls.Type.supertype(type1._type, type2._type, cls.common_supertypes)
            if t is not None:
                return cls.make_type(t)
            return None

        if type1._type == type2._type:
            return cls._supertype(type1, type2)

        return None

    @classmethod
    @abc.abstractmethod
    def _supertype(cls, type1: ColumnType, type2: ColumnType) -> Optional[ColumnType]:
        """
        Class-specific implementation of determining the supertype. type1 and type2 are from the same subclass of
        ColumnType.
        """
        pass

    @classmethod
    def infer_literal_type(cls, val: Any) -> Optional[ColumnType]:
        if isinstance(val, str):
            return StringType()
        if isinstance(val, int):
            return IntType()
        if isinstance(val, float):
            return FloatType()
        if isinstance(val, bool):
            return BoolType()
        if isinstance(val, datetime.datetime) or isinstance(val, datetime.date):
            return TimestampType()
        if isinstance(val, np.ndarray):
            col_type = ArrayType.from_literal(val)
            if col_type is not None:
                return col_type
            # this could still be json-serializable
        if isinstance(val, dict) or isinstance(val, np.ndarray):
            try:
                JsonType().validate_literal(val)
                return JsonType()
            except TypeError:
                return None
        return None


    @classmethod
    def from_python_type(cls, t: type) -> Optional[ColumnType]:
        if t in _python_type_to_column_type:
            return _python_type_to_column_type[t]
        elif isinstance(t, typing._UnionGenericAlias) and t.__args__[1] is type(None):
            # `t` is a type of the form Optional[T] (equivalently, Union[T, None]).
            # We treat it as the underlying type but with nullable=True.
            if t.__args__[0] in _python_type_to_column_type:
                underlying = copy(_python_type_to_column_type[t.__args__[0]])
                underlying.nullable = True
                return underlying

        return None


    def validate_literal(self, val: Any) -> None:
        """Raise TypeError if val is not a valid literal for this type"""
        if val is None:
            if not self.nullable:
                raise TypeError('Expected non-None value')
            else:
                return
        self._validate_literal(val)

    def validate_media(self, val: Any) -> None:
        """
        Raise TypeError if val is not a path to a valid media file (or a valid in-memory byte sequence) for this type
        """
        if self.is_media_type():
            raise NotImplementedError(f'validate_media() not implemented for {self.__class__.__name__}')

    def _validate_file_path(self, val: Any) -> None:
        """Raises TypeError if not a valid local file path or not a path/byte sequence"""
        if isinstance(val, str):
            parsed = urllib.parse.urlparse(val)
            if parsed.scheme != '' and parsed.scheme != 'file':
                return
            path = Path(urllib.parse.unquote(parsed.path))
            if not path.is_file():
                raise TypeError(f'File not found: {str(path)}')
        else:
            if not isinstance(val, bytes):
                raise TypeError(f'expected file path or bytes, got {type(val)}')

    @abc.abstractmethod
    def _validate_literal(self, val: Any) -> None:
        """Raise TypeError if val is not a valid literal for this type"""
        pass

    @abc.abstractmethod
    def _create_literal(self, val : Any) -> Any:
        """Create a literal of this type from val, including any needed conversions.
             val is guaranteed to be non-None"""
        return val

    def create_literal(self, val: Any) -> Any:
        """Create a literal of this type from val or raise TypeError if not possible"""
        if val is not None:
            val = self._create_literal(val)

        self.validate_literal(val)
        return val

    def print_value(self, val: Any) -> str:
        return str(val)

    def is_scalar_type(self) -> bool:
        return self._type in self.scalar_types

    def is_numeric_type(self) -> bool:
        return self._type in self.numeric_types

    def is_invalid_type(self) -> bool:
        return self._type == self.Type.INVALID

    def is_string_type(self) -> bool:
        return self._type == self.Type.STRING

    def is_int_type(self) -> bool:
        return self._type == self.Type.INT

    def is_float_type(self) -> bool:
        return self._type == self.Type.FLOAT

    def is_bool_type(self) -> bool:
        return self._type == self.Type.BOOL

    def is_timestamp_type(self) -> bool:
        return self._type == self.Type.TIMESTAMP

    def is_json_type(self) -> bool:
        return self._type == self.Type.JSON

    def is_array_type(self) -> bool:
        return self._type == self.Type.ARRAY

    def is_image_type(self) -> bool:
        return self._type == self.Type.IMAGE

    def is_video_type(self) -> bool:
        return self._type == self.Type.VIDEO

    def is_audio_type(self) -> bool:
        return self._type == self.Type.AUDIO

    def is_document_type(self) -> bool:
        return self._type == self.Type.DOCUMENT

    def is_media_type(self) -> bool:
        # types that refer to external media files
        return self.is_image_type() or self.is_video_type() or self.is_audio_type() or self.is_document_type()

    @abc.abstractmethod
    def to_sql(self) -> str:
        """
        Return corresponding Postgres type.
        """
        pass

    @abc.abstractmethod
    def to_sa_type(self) -> Any:
        """
        Return corresponding SQLAlchemy type.
        return type Any: there doesn't appear to be a superclass for the sqlalchemy types
        """
        assert self._type != self.Type.INVALID
        if self._type == self.Type.STRING:
            return sql.String
        if self._type == self.Type.INT:
            return sql.Integer
        if self._type == self.Type.FLOAT:
            return sql.Float
        if self._type == self.Type.BOOL:
            return sql.Boolean
        if self._type == self.Type.TIMESTAMP:
            return sql.TIMESTAMP
        if self._type == self.Type.IMAGE:
            # the URL
            return sql.String
        if self._type == self.Type.JSON:
            return sql.dialects.postgresql.JSONB
        if self._type == self.Type.ARRAY:
            return sql.VARBINARY
        assert False

    @abc.abstractmethod
    def to_arrow_type(self) -> 'pyarrow.DataType':
        assert False, f'Have not implemented {self.__class__.__name__} to Arrow'
 
    @staticmethod
    def no_conversion(v: Any) -> Any:
        """
        Special return value of conversion_fn() that indicates that no conversion is necessary.
        Should not be called
        """
        assert False

    def conversion_fn(self, target: ColumnType) -> Optional[Callable[[Any], Any]]:
        """
        Return Callable that converts a column value of type self to a value of type 'target'.
        Returns None if conversion isn't possible.
        """
        return None


class InvalidType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.INVALID, nullable=nullable)

    def to_sql(self) -> str:
        assert False

    def to_sa_type(self) -> Any:
        assert False

    def to_arrow_type(self) -> 'pyarrow.DataType':
        assert False

    def print_value(self, val: Any) -> str:
        assert False

    def _validate_literal(self, val: Any) -> None:
        assert False

class StringType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.STRING, nullable=nullable)

    def conversion_fn(self, target: ColumnType) -> Optional[Callable[[Any], Any]]:
        if not target.is_timestamp_type():
            return None
        def convert(val: str) -> Optional[datetime]:
            try:
                dt = datetime.datetime.fromisoformat(val)
                return dt
            except ValueError:
                return None
        return convert

    def to_sql(self) -> str:
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        return pa.string()

    def print_value(self, val: Any) -> str:
        return f"'{val}'"

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, str):
            raise TypeError(f'Expected string, got {val.__class__.__name__}')


class IntType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.INT, nullable=nullable)

    def to_sql(self) -> str:
        return 'BIGINT'

    def to_sa_type(self) -> str:
        return sql.BigInteger
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        return pa.int64() # to be consistent with bigint above

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, int):
            raise TypeError(f'Expected int, got {val.__class__.__name__}')


class FloatType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.FLOAT, nullable=nullable)

    def to_sql(self) -> str:
        return 'FLOAT'

    def to_sa_type(self) -> str:
        return sql.Float
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa
        return pa.float32()

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, float):
            raise TypeError(f'Expected float, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, int):
            return float(val)
        return val

class BoolType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.BOOL, nullable=nullable)

    def to_sql(self) -> str:
        return 'BOOLEAN'

    def to_sa_type(self) -> str:
        return sql.Boolean
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        return pa.bool_()

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, bool):
            raise TypeError(f'Expected bool, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, int):
            return bool(val)
        return val

class TimestampType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.TIMESTAMP, nullable=nullable)

    def to_sql(self) -> str:
        return 'INTEGER'

    def to_sa_type(self) -> str:
        return sql.TIMESTAMP
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        return pa.timestamp('us') # postgres timestamp is microseconds

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, datetime.datetime) and not isinstance(val, datetime.date):
            raise TypeError(f'Expected datetime.datetime or datetime.date, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, str):
            return datetime.datetime.fromisoformat(val)
        return val

class JsonType(ColumnType):
    # TODO: type_spec also needs to be able to express lists
    def __init__(self, type_spec: Optional[Dict[str, ColumnType]] = None, nullable: bool = False):
        super().__init__(self.Type.JSON, nullable=nullable)
        self.type_spec = type_spec

    def _as_dict(self) -> Dict:
        result = super()._as_dict()
        if self.type_spec is not None:
            type_spec_dict = {field_name: field_type.serialize() for field_name, field_type in self.type_spec.items()}
            result.update({'type_spec': type_spec_dict})
        return result

    @classmethod
    def _from_dict(cls, d: Dict) -> ColumnType:
        type_spec = None
        if 'type_spec' in d:
            type_spec = {
                field_name: cls.deserialize(field_type_dict) for field_name, field_type_dict in d['type_spec'].items()
            }
        return cls(type_spec, nullable=d['nullable'])

    def to_sql(self) -> str:
        return 'JSONB'

    def to_sa_type(self) -> str:
        return sql.dialects.postgresql.JSONB
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        return pa.string() # TODO: weight advantage of pa.struct type.

    def print_value(self, val: Any) -> str:
        val_type = self.infer_literal_type(val)
        if val_type == self:
            return str(val)
        return val_type.print_value(val)

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, dict) and not isinstance(val, list):
            raise TypeError(f'Expected dict or list, got {val.__class__.__name__}')
        try:
            _ = json.dumps(val)
        except TypeError as e:
            raise TypeError(f'Expected JSON-serializable object, got {val}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, tuple):
            val = list(val)
        return val

class ArrayType(ColumnType):
    def __init__(
            self, shape: Tuple[Union[int, None], ...], dtype: ColumnType, nullable: bool = False):
        super().__init__(self.Type.ARRAY, nullable=nullable)
        self.shape = shape
        assert dtype.is_int_type() or dtype.is_float_type() or dtype.is_bool_type() or dtype.is_string_type()
        self.dtype = dtype._type

    def _supertype(cls, type1: ArrayType, type2: ArrayType) -> Optional[ArrayType]:
        if len(type1.shape) != len(type2.shape):
            return None
        base_type = ColumnType.supertype(type1.dtype, type2.dtype)
        if base_type is None:
            return None
        shape = [n1 if n1 == n2 else None for n1, n2 in zip(type1.shape, type2.shape)]
        return ArrayType(tuple(shape), base_type)

    def _as_dict(self) -> Dict:
        result = super()._as_dict()
        result.update(shape=list(self.shape), dtype=self.dtype.value)
        return result

    def __str__(self) -> str:
        return f'{self._type.name.lower()}({self.shape}, dtype={self.dtype.name})'

    @classmethod
    def _from_dict(cls, d: Dict) -> ColumnType:
        assert 'shape' in d
        assert 'dtype' in d
        shape = tuple(d['shape'])
        dtype = cls.make_type(cls.Type(d['dtype']))
        return cls(shape, dtype, nullable=d['nullable'])

    @classmethod
    def from_literal(cls, val: np.ndarray) -> Optional[ArrayType]:
        # determine our dtype
        assert isinstance(val, np.ndarray)
        if np.issubdtype(val.dtype, np.integer):
            dtype = IntType()
        elif np.issubdtype(val.dtype, np.floating):
            dtype = FloatType()
        elif val.dtype == np.bool_:
            dtype = BoolType()
        elif val.dtype == np.str_:
            dtype = StringType()
        else:
            return None
        return cls(val.shape, dtype=dtype, nullable=True)

    def is_valid_literal(self, val: np.ndarray) -> bool:
        if not isinstance(val, np.ndarray):
            return False
        if len(val.shape) != len(self.shape):
            return False
        # check that the shapes are compatible
        for n1, n2 in zip(val.shape, self.shape):
            if n1 is None:
                return False
            if n2 is None:
                # wildcard
                continue
            if n1 != n2:
                return False
        return val.dtype == self.numpy_dtype()

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, np.ndarray):
            raise TypeError(f'Expected numpy.ndarray, got {val.__class__.__name__}')
        if not self.is_valid_literal(val):
            raise TypeError((
                f'Expected ndarray({self.shape}, dtype={self.numpy_dtype()}), '
                f'got ndarray({val.shape}, dtype={val.dtype})'))

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, (list,tuple)):
            return np.array(val)
        return val

    def to_sql(self) -> str:
        return 'BYTEA'

    def to_sa_type(self) -> str:
        return sql.LargeBinary
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        if any([n is None for n in self.shape]):
            raise TypeError(f'Cannot convert array with unknown shape to Arrow')        
        return pa.fixed_shape_tensor(pa.from_numpy_dtype(self.numpy_dtype()), self.shape)

    def numpy_dtype(self) -> np.dtype:
        if self.dtype == self.Type.INT:
            return np.dtype(np.int64)
        if self.dtype == self.Type.FLOAT:
            return np.dtype(np.float32)
        if self.dtype == self.Type.BOOL:
            return np.dtype(np.bool_)
        if self.dtype == self.Type.STRING:
            return np.dtype(np.str_)
        assert False


class ImageType(ColumnType):
    def __init__(
            self, width: Optional[int] = None, height: Optional[int] = None, size: Optional[Tuple[int, int]] = None,
            mode: Optional[str] = None, nullable: bool = False
    ):
        """
        TODO: does it make sense to specify only width or height?
        """
        super().__init__(self.Type.IMAGE, nullable=nullable)
        assert not(width is not None and size is not None)
        assert not(height is not None and size is not None)
        if size is not None:
            self.width = size[0]
            self.height = size[1]
        else:
            self.width = width
            self.height = height
        self.mode = mode

    def __str__(self) -> str:
        if self.width is not None or self.height is not None or self.mode is not None:
            params_str = ''
            if self.width is not None:
                params_str = f'width={self.width}'
            if self.height is not None:
                if len(params_str) > 0:
                    params_str += ', '
                params_str += f'height={self.height}'
            if self.mode is not None:
                if len(params_str) > 0:
                    params_str += ', '
                params_str += f'mode={self.mode}'
            params_str = f'({params_str})'
        else:
            params_str = ''
        return f'{self._type.name.lower()}{params_str}'

    def _is_supertype_of(self, other: ImageType) -> bool:
        if self.mode != other.mode:
            return False
        if self.width is None and self.height is None:
            return True
        if self.width != other.width and self.height != other.height:
            return False

    @property
    def size(self) -> Optional[Tuple[int, int]]:
        if self.width is None or self.height is None:
            return None
        return (self.width, self.height)

    @property
    def num_channels(self) -> Optional[int]:
        return None if self.mode is None else self.mode.num_channels()

    def _as_dict(self) -> Dict:
        result = super()._as_dict()
        result.update(width=self.width, height=self.height, mode=self.mode)
        return result

    @classmethod
    def _from_dict(cls, d: Dict) -> ColumnType:
        assert 'width' in d
        assert 'height' in d
        assert 'mode' in d
        return cls(width=d['width'], height=d['height'], mode=d['mode'], nullable=d['nullable'])

    def conversion_fn(self, target: ColumnType) -> Optional[Callable[[Any], Any]]:
        if not target.is_image_type():
            return None
        assert isinstance(target, ImageType)
        if (target.width is None) != (target.height is None):
            # we can't resize only one dimension
            return None
        if (target.width == self.width or target.width is None) \
            and (target.height == self.height or target.height is None) \
            and (target.mode == self.mode or target.mode is None):
            # nothing to do
            return self.no_conversion
        def convert(img: PIL.Image.Image) -> PIL.Image.Image:
            if self.width != target.width or self.height != target.height:
                img = img.resize((target.width, target.height))
            if self.mode != target.mode:
                img = img.convert(target.mode.to_pil())
            return img
        return convert

    def to_sql(self) -> str:
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        return pa.binary()

    def _validate_literal(self, val: Any) -> None:
        if isinstance(val, PIL.Image.Image):
            return
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        assert isinstance(val, str)
        try:
            _ = PIL.Image.open(val)
        except PIL.UnidentifiedImageError:
            raise excs.Error(f'Not a valid image: {val}') from None

class VideoType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.VIDEO, nullable=nullable)

    def to_sql(self) -> str:
        # stored as a file path
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String
    
    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa # pylint: disable=import-outside-toplevel
        return pa.string()

    def _validate_literal(self, val: Any) -> None:
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        assert isinstance(val, str)
        try:
            with av.open(val, 'r') as fh:
                if len(fh.streams.video) == 0:
                    raise excs.Error(f'Not a valid video: {val}')
                # decode a few frames to make sure it's playable
                # TODO: decode all frames? but that's very slow
                num_decoded = 0
                for frame in fh.decode(video=0):
                    _ = frame.to_image()
                    num_decoded += 1
                    if num_decoded == 10:
                        break
                if num_decoded < 2:
                    # this is most likely an image file
                    raise excs.Error(f'Not a valid video: {val}')
        except av.AVError:
            raise excs.Error(f'Not a valid video: {val}') from None

class AudioType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.AUDIO, nullable=nullable)

    def to_sql(self) -> str:
        # stored as a file path
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String

    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa  # pylint: disable=import-outside-toplevel
        return pa.string()

    def _validate_literal(self, val: Any) -> None:
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        try:
            with av.open(val) as container:
                if len(container.streams.audio) == 0:
                    raise excs.Error(f'No audio stream in file: {val}')
                audio_stream = container.streams.audio[0]

                # decode everything to make sure it's playable
                # TODO: is there some way to verify it's a playable audio file other than decoding all of it?
                for packet in container.demux(audio_stream):
                    for _ in packet.decode():
                        pass
        except av.AVError as e:
            raise excs.Error(f'Not a valid audio file: {val}\n{e}') from None

class DocumentType(ColumnType):
    @enum.unique
    class DocumentFormat(enum.Enum):
        HTML = 0
        MD = 1
        PDF = 2

    def __init__(self, nullable: bool = False, doc_formats: Optional[str] = None):
        super().__init__(self.Type.DOCUMENT, nullable=nullable)
        if doc_formats is not None:
            type_strs = doc_formats.split(',')
            for type_str in type_strs:
                if not hasattr(self.DocumentFormat, type_str):
                    raise ValueError(f'Invalid document type: {type_str}')
            self._doc_formats = [self.DocumentFormat[type_str.upper()] for type_str in type_strs]
        else:
            self._doc_formats = [t for t in self.DocumentFormat]

    def to_sql(self) -> str:
        # stored as a file path
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String

    def to_arrow_type(self) -> 'pyarrow.DataType':
        import pyarrow as pa  # pylint: disable=import-outside-toplevel
        return pa.string()

    def _validate_literal(self, val: Any) -> None:
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        assert isinstance(val, str)
        from pixeltable.utils.documents import get_document_handle
        with open(val, 'r') as fh:
            try:
                s = fh.read()
                dh = get_document_handle(s)
                if dh is None:
                    raise excs.Error(f'Not a recognized document format: {val}')
            except Exception as e:
                raise excs.Error(f'Not a recognized document format: {val}') from None


# A dictionary mapping various Python types to their respective ColumnTypes.
# This can be used to infer Pixeltable ColumnTypes from type hints on Python
# functions. (Since Python functions do not necessarily have type hints, this
# should always be an optional/convenience inference.)
_python_type_to_column_type: dict[type, ColumnType] = {
    str: StringType(),
    int: IntType(),
    float: FloatType(),
    bool: BoolType(),
    datetime.datetime: TimestampType(),
    datetime.date: TimestampType(),
    list: JsonType(),
    dict: JsonType(),
    PIL.Image.Image: ImageType()
}
